/*
    base and operators defining the Hilbert space
    base = spin-configurations,
    p.e. get the index from config or reverse, get symetric conf.
    - look at block_no
  
 */
#include "hilbert.h"
#include "mem.h"
#include "tbase.h"
#include "ansatz.h"
#include "array_m.h"
#include "block.h"       /* mpi_* declarations */
#ifdef CONFIG_PTHREAD
#include <pthread.h>
#endif
#ifdef CONFIG_MPI
#include <mpi.h>
#endif

ULLong  npa[NN+1][NN+1][NN+1];   /* (o+u+d)!/(o!u!d!) stored as 3D table */

zahl fak(int i){zahl r=1.0;int j;for(j=2;j<=i;j++)r*=j;return(r);}
/* ------------------------------------------------------
 *  calculate the binominal coefficient (use pascal triangle(?))
 *  over(n,k) = n!/((n-k)!k!) = over(n-1,k-1)*n/k = over(n-1,k)+over(n-1,r-1)
 * --------------------------------------------------------- */
ULLong over(int n,int k){
  static ULLong pascal[2*NN+1][2*NN+1]; static int ok=0; int i,j;
  if(!ok){ ok=1;		/* de: Pascal-Dreieck bauen */
    for(i=0;i<=2*NN;i++)for(j=0;j<=2*NN;j++)pascal[i][j]=(ULLong)((j)?0l:1l);
    for(i=1;i<=2*NN;i++)for(j=1;j<=i;j++)
      pascal[i][j]=pascal[i-1][j]+pascal[i-1][j-1];	/* OVL ? */
  }  
  if(n>2*NN || k>n) return 0l;		/* NN => 2*NN  Nov98 JS */ 
  return pascal[n][k]; /* only 0<=k<=n<=NN */    
}


#if CM(1,1,0)  
/* Perm. mit Wdhlg (3 Elemente)=(o+u+d)!/(o!u!d!)=over(n-u,d)*over(n,u) */
 ULLong np(int o,int u,int d){ return(over(o+u+d,u)*over(o+d,d));}
#else	/* Produkt 2er Kombinat. (2 Elemente)=(n over u)(n over d) */
 ULLong np(int n,int u,int d){return(over(n,u)*over(n,d));}
#endif
void   paini(){ int a,b,c;
    for(a=0;a<=NN;a++)for(b=0;b<=NN;b++)
    for(c=0;c<=NN;c++)npa[a][b][c]=np(a,b,c);
    return;}
    
/* number of spin configurations (nn,nu,nd), same as n1 without Sym */
ULLong numconf(){ return(npa[nn-CM(1,1,0)*(nu+nd)][nu][nd]);}

/* --------------------------------------------------------------------*
 *                    in-out-functions
 * call this function not more than 8 times within one printf() :) 
 *  v2.26:  .=empty-site, u=up, d=down, 3=up+down 
 *  v2.27:  .=empty-site, o=up, x=down, Q=up+down (looks better)
 *          think on pointers pointing to you "o" or away from you "x"
 * ToDo: for S>1/2 output for tJ,tU (s,m=-s...+s) and for JJ (m)=(-s...+s)
 * - not thread safe
 * ------------------------------------------------------------------- */
char *ccfg(tbase ic){ /* Spinconfiguration |543210> zB: |o.xoxo> */
  static char s1[8*(NN+1)], s2[8]=".oxQ"; char *s3;
  static int x=0; int i;
  x=(x+1)%8; s3=s1+x*(NN+1);
  for (i=nn-1;i>=0;i--) s3[nn-1-i]=s2[ geti(ic,i) ];
  s3[nn]='\0';
  return(s3);
}

/* ----------------------------------------------------------------------- * 
 * symmetry-OP   PermOp(ip)*ib1=s*ib2 , s = +1/-1  ip = permutation vector 
 * ----------------------------------------------------------------------- */
/*   example:
 P = [0 2 1 3] = (0) (1,2) (3)
 ib1= (.. du .u d.) = sites (3 2 1 0) = bits (76 54 32 10)
 m1=7=3d -
 m1=6=3u -						 
 m1=5=2d m2=3=1d i2=(.. .. d. ..) s=0       i3=(11 11 10 00)=(0000,1110)
 m1=4=2u m2=2=1u i2=(.. .. du ..) s^=i3_1=0 i3=(00 00 01 00)=(1110,1110)
 m1=3=1d -
 m1=2=1u m2=4=2u i2=(.. .u du ..) s^=i3_3=0 i3=(11 11 01 00)=(0010,1110)
 m1=1=0d m2=1=0d i2=(.. .u du d.) s^=i3_0=0 i3=(00 00 10 10)=(0010,0001)
 m1=0=0u -      
 ------------------------------------------------------------------------ */
/* ms!=0 multiply phase with k, ToDo: remove and simplify
 * later build in FPGA!
 */
/* static inline */ /* check asm-code for inlining gcc -S -dA */
/* tphase is complex in complex mode, ToDo: always signed char? +/-1 */
void get_symconfig(tbase ib1,int ip,tbase *ib2,tphase *ss,int ms){
  register tbase i2;
  register int m1, m2;
#if CM(0,1,1)
  register tbase i3; int s;
#endif
  *ib2=ib1; *ss=+1; if(ip<0) return; /* 1-Operator (0,identity) */
#if CM(0,1,1)
  i3=i2=0l; s=0;				/* remove from NN-1 to 0 */ 
  for(m1=2*nn-1;m1>=0;m1--)if((ib1>>m1)&1){	/* m1 = old index */
    m2 = sym.i[ip][m1/2]*2+(m1&1);		/* m2 = new index */
    i2 |= ((tbase)1) << m2;			/* move spin to site m2 */
    if (m2>0) s^=(int)(1&(i3>>(m2-1))); 	/* number of spins jumped over */ 
    i3 ^= (~(tbase)0)<<m2;		/* derived fermionic sign */
  }
  (*ss) *= 1-2*(s&1);   /* ToDo: return s in units of I*2Pi/2 */
#else /* CM(1,0,0) Heisenberg, ToDo: make speedtests over all n1 */ 
  i2=0l;
#if 1  /* how much asm commands? */
  for (m1=0; ib1!=0 && m1<nn; m1++, ib1>>=1) if (1&ib1) { /* faster if nd<=nu */
    m2 = sym.i[ip][m1];
    i2 |= ((tbase)1) << m2; /* ToDo: m2>32 failes on 32bit gcc! use table[NN] */
  }	/* ToDo: uebergabe &sym.i[ip] statt ip ???  */
#else
  t_sites *a=sym.i[ip];
  for (m1=NN; ; ) { /* faster if nd<=nu */
    i2 |= ((tbase)(1&ib1)) << *a;      /* asm: jmp if NN>sizeof(int) */
    a++; m1--; ib1>>=1; if (!m1 || !ib1) break;
  }	/* ToDo: uebergabe &sym.i[ip] statt ip ???  */
#endif
#endif /* JJ */
  *ib2=i2;
  if(ms){
    (*ss)*=sym.w[ip];   /* better add fractional angles (integer) outside */
#if VecType & 4 /* complex phase */
    /* remove rounding errors, do it outside */
    if(fabs(imag(*ss))<1e-9) (*ss)=C(real(*ss),0);
#endif
  }  
  return;
}

tbase m55=0;    /* dynamic initialisation only in C++ => initmasks */
tbase mAA=0;
tbase m0F=0;

/* initialize bitmasks m55,mAA,m0F */
int initmasks(int nn){
  int i; tbase r=0;
  for(i=0;(unsigned)i<sizeof(tbase)*8;i+=2) r|=((tbase)1)<<i;
  m55=r;
  mAA=r<<1;
  m0F=~(((~(tbase)0)<<(nn-1))<<1); /* (int32)<<32 won't work on x86 */
  /*  printf(" ini55: %s\n",ccfg(r)); */
  return 0;
}
    
/* ----------------------------------------------------------------------- * 
 * Sz-symmetry (spin inversion) u<->d,   attention: ss may point to s1 !!!
 *  ToDo: change type of tphase from complex value (k*i*2pi/l) to integer angle (k)
 *  ToDo: dont input s1, better return the sign
 *  used in xval.c wop_ud() if nu==nd obsoletem and in hilbert.c
 * ----------------------------------------------------------------------- */
static inline
void get_sudconfig(tbase ib1,tphase s1,tbase *ib2,tphase *ss){
  /* ib2(ss) can also point to ib1(s1) !!! */
  *ib2=ib1;*ss=s1;		/* default: 1-operator */
  if(nu!=nd || !sym.wud) return; /* ToDo: remove it, decide it outside! */
#if CM(1,0,0)
  (*ib2)=ib1 ^ m0F;
#else /* tJ + tU */
  (*ib2)=((mAA & ib1)>>1) | ((m55 & ib1)<<1);
#if CM(0,0,1)  /* 3=(u+d) will change sign, ToDo: only if sign needed? slow */
  { tbase1 i3;
    for (i3=((tbase)3)<<(2*nn-2); i3!=0; i3>>=2) if ((ib1&i3)==i3) (*ss)*=-1;
  }
#endif /* tU only */
#endif /* tJ + tU */
  (*ss)*=sym.wud;	/* ToDo: can be done outside, return only +/-1 */
  return;
}

/* --------------------------------------------------------------------	*
 * 		Spin-Konfigurationen ki und deren Index i		*
 * --------------------------------------------------------------------	*/
#define MF(x)   ( ~((~(tbase)0)<<(x)))	/* << only int/long, not longlong */
#define MA(x)   (mAA & MF(x))		/* MF(x) = x * 1 bits */
#define M5(x)   (m55 & MF(x))
#define M1(x)   (  ((tbase)  1)<<(x) )
#define M2(x)   (  ((tbase)  2)<<(x) )
#define M3(x)   (  ((tbase)  3)<<(x) )
#define ME(x)   (  ((tbase)0-2)<<(x) )
#define MC(x)   (  ((tbase)0-4)<<(x) )
#if CM(0,1,0)
inline tbase next(tbase cfg){                                       /* ki++ */ 
  int i=0,d=0,u=0; 	/* Achtung: ki<<32 bei 32bit nicht definiert*/
  if(cfg&2l)d+=2; else if(cfg&1l)u+=2;
  while((i+=2)<2*nn)switch((int)((cfg>>i)&3l)){
   case 0:       if(u){ u-=2;
     return((cfg & MC(i)) | M1(i) | M5(u)<<d | MA(d));} else if(d){ d-=2;
     return((cfg & MC(i)) | M2(i) | M5(u)<<d | MA(d));} break;
   case 1: u+=2; if(d){ d-=2;
     return((cfg & MC(i)) | M2(i) | M5(u)<<d | MA(d));} break;
   case 2: d+=2;}
  return(M5(u)<<d | MA(d));}

tbase  ll1(ULLong ii){ /* getbasis from index 0..N1-1 */
  ULLong i,t;tbase j;int io,iu,id;
  j=0l;io=nn-nu-nd;iu=nu;id=nd;i=ii+1l;
  for(;;){
   if(io>0){ if(i<=(t=npa[io-1][iu][id])){j=(j<<2);  io--;continue;}i-=t;}
   if(iu>0){ if(i<=(t=npa[io][iu-1][id])){j=(j<<2)|1;iu--;continue;}i-=t;}
   if(id>0){ if(i<=(t=npa[io][iu][id-1])){j=(j<<2)|2;id--;continue;}i-=t;}
   else break;}
  return(j);}
    
ULLong ll2(tbase ii){ /* n-dimensionale Linkodierung */
/* Inp: Spinconfiguration, Out: Index des Basiszustandes */
    ULLong j=0l;int io,iu,id,i,i1;
    io=nn-nu-nd;iu=nu;id=nd;for(i=nn-1;i>=0;i--){
      i1=(int)(3&(ii>>(2*i)));
      if(i1>0 && io>0)j+=npa[io-1][iu][id];
      if(i1>1 && iu>0)j+=npa[io][iu-1][id];
      if(i1==0)io--;
      if(i1==1)iu--;
      if(i1==2)id--;}
    return(j);}

#elif CM(1,0,0)

/* 01101 i=1 d=1 inp  */
/* 01110 i=2 d=1 !!!  */
/* 10011 i=2 d=2      */
/* 10101 i=1 d=1      */
/* 10110 i=2 d=1      */
/* 12233=123=LM
   00011 002 i=2 d=1
   00101 011
   01100 020
   10001 101
   10100 110
 */
/* last change: long long Anpassung 9.4.1998 JS */
/* MP-version Dec99, better readable */
#if defined(CONFIG_S1SYM)
/* improved next()-function, some easy symmetries included */
static inline tbase next(tbase cfg){	/* ??? */ 
  int i,d,k,j,v=0; /* i - position of first "01"-bit sequence from bit0 */
  i=d=0;           /* d - number of 1-bits exclusive 01 pattern */
  if(sym.num_lm!=nn){   /* ToDo: replace by S(+,i)-ops and s>1/2 sites */
    if(v&1)printf("\n0 i=%2d d=%2d next=%s",i,d,ccfg(cfg));
    for(;i!=nn;i++){ /* searching 01 bit pattern */
     if(1&(cfg>>i)){ /* clear spin, if no higher LM-spin */
       for(k=i+1;k<nn;k++)if(sym.lm[k]==sym.lm[i])if(1&(cfg>>k))break;
       if(k==nn)
       for(k=0  ;k<=i;k++)if(sym.lm[k]==sym.lm[i])if(1&(cfg>>k)) 
       { d++;cfg&=~M1(k); }
     }
     else { /* OVL to other LM */
       for(j=0,k=0;k<i;k++)if(sym.lm[k]==sym.lm[i])j++;
       if(d>j)break;
     }
    }
    if(i<nn) { cfg |= M1(i); d--; } else i=d;
    /* first spread spins over different LM cluster */
    if(v&2)printf("\n1 i=%2d d=%2d next=%s",i,d,ccfg(cfg));
    for(k=0;k<i && d>0;k++)if(sym.lm[k]==sym.lm[i]){ d--;cfg|=M1(k); }
    /* now spread over the rest */
    if(v&4)printf("\n2 i=%2d d=%2d next=%s",i,d,ccfg(cfg));
    for(k=0;k<i && d>0;k++)if(!((cfg>>k)&1)){ d--;cfg|=M1(k); }
    if(v&8)printf("\n3 i=%2d d=%2d next=%s",i,d,ccfg(cfg));
    return(cfg);
  }
  for(;i!=nn;i++){ if((cfg>>i)&1)d++;else if(d)break; }
  return ( (cfg & ~MF(i)) | M1(i) | MF(--d) ); /* d==0: x<<-1 ???  */
}
#else
static inline tbase next(tbase cfg){	/* get next code (nd of N) */ 
  int i,d;	/* i - position of first "01"-bit sequence from bit0 */
  i=d=0;	/* d - number of 1-bits exclusive 01 pattern */
  for(;i!=nn;i++){ if((cfg>>i)&1)d++;else if(d)break; }
  return ( (cfg & ~MF(i)) | M1(i) | MF(--d) ); /* d==0: x<<-1 ???  */
}
#endif

tbase  ll1(ULLong ii){	/* getbasis from index 0..N1-1 */
  ULLong i,t;tbase j;int iu,id;
  j=0l;iu=nu;id=nd;i=ii+1l;if(nn!=nu+nd)F0("nu+nd != NN")
  for(;;){
    if(iu>0){ if(i<=(t=npa[0][iu-1][id])){j=(j<<1)|0;iu--;continue;}i-=t;}
    /* else { j=(j<<id)|MF(id);break; F A S T E R ???? } */
    if(id>0){ if(i<=(t=npa[0][iu][id-1])){j=(j<<1)|1;id--;continue;}i-=t;}
    break;}
  return(j);
}
    
ULLong ll2(tbase ii){	/* n-dimensional Lin-coding */
/* Inp: spin configuration, Out: index of base state */
  ULLong j=0l; int iu,id,i,i1;
  iu=nu; id=nd; if(nn!=nu+nd) F0("nu+nd != NN")
  for(i=nn-1;i>=0;i--){
    i1=(int)(1&(ii>>i));
    if (i1>0 && iu>0 && id>0) j+=npa[0][iu-1][id];
    if (i1==0) iu--;
    if (i1==1) id--;}
  return(j);
}

#elif CM(0,0,1)  /* now the coding for the Hubbard model */

inline tbase next(tbase cfg){  				/* ki++ */ 
  int i,d,u;if(cfg&1l)u=2;else u=0;if(cfg&2l)d=2;else d=0;  
  for(i=2;i<2*nn;i+=2)switch((int)(3&(cfg>>i))){      /* -1<<2=-4 */
   case 0:
    if(u){ return((cfg & MC(i)) | M1(i) | M5(u-2) | MA(d));} else 
    if(d){ return((cfg & MC(i)) | M2(i) | M5(u) | MA(d-2));} break;
   case 1:  u+=2; 
    if(d){ return((cfg & MC(i)) | M2(i) | M5(u) | MA(d-2));} break;
   case 2:
    if(u){ return((cfg & MC(i)) | M3(i) | M5(u-2) | MA(d));} d+=2; break;
   case 3:  u+=2;d+=2;}
  return( M5(u) | MA(d) );}

tbase  ll1(ULLong ii){	/* get configuration from index 0..N1-1 */
    tbase j;int in,iu,id;ULLong i,t;i=ii;j=0l;iu=nu;id=nd;
    for(in=nn-1;in>=0;in--){
     if(in>=iu && in>=id){
         if(i<(t=npa[in][iu  ][id  ])){continue;}i-=t;}
     if(iu>0 && in>=id){
         if(i<(t=npa[in][iu-1][id  ])){j|=M1(2*in);iu--;continue;}i-=t;}
     if(id>0 && in>=iu){
         if(i<(t=npa[in][iu  ][id-1])){j|=M2(2*in);id--;continue;}i-=t;}
     if(id>0 && iu>0)
         if(i<   npa[in][iu-1][id-1] ){j|=M3(2*in);id--;iu--;continue;}
     break;}
    return(j);}
    
ULLong ll2(tbase ii){                /* ki -> i */
    ULLong j;int iu,id,i,i1;j=0l;iu=nu;id=nd;
    for(i=nn-1;i;i--){ if(i<iu)if(i<id)break;i1=3&(ii>>(i<<1));
      if(i1>0){  if(i>=iu   && i>=id        )j+=npa[i][iu  ][id];
       if(i1>1){ if(i>=iu-1 && i>=id && iu>0)j+=npa[i][iu-1][id];id--;
        if(i1>2){if(i>=iu   && i>=id        )j+=npa[i][iu  ][id];iu--;}}
       else iu--;}}
    return(j);}
#endif
#undef MF
#undef MA
#undef M5
#undef M1
#undef M2
#undef M3
#undef MC
#undef ME

/* suche entry in sorted tab[0..l-1], not found => 0 */
/* ToDo: it should be easier with uuu and ddd-tab if tU */
ULong findcfg(tbase entry, ULong l){
  ULong j,m;j=0l;  /* v203 tbase1=>ULong, 98: ULong=>tbase1 */
#if (CM(1,2,2)*NN-1)>8*S_long-1
  m=~(~(ULong)0u>>1);		/* set msb (highest bit) */
#else
  m=((ULong)1)<<(CM(1,2,2)*nn-1);
  if(m==0 || m==1)F0("OVL");
#endif
  while(m){if((j|m)<l)	/* long unsigned == unsigned long ? */
    if(getl1(j|m)<=entry) j|=m;
    m>>=1;
  }
  /* if(getl1(j)!=entry)j=0l;			 not found */
  return((ULong) j);			/* ULong/tbase1 ??? */
}

/* ------------------------ symmetric base config (class Hbase) ----
 * inp: sym.k[], sym.i[] 
 * out: |rc)=sign(s)*|s|^(-1/2)*|rc],  s=<ib1|rc)
 * |] is sym-base, summe aller zu |ib1> symmetr. configs 
 * |) is normed sym-base, 
 * |> is spin-configuration (1st base) 
 * ------------------------------------------------------------- */
 
/*
 *  print symmetric base config for debugging purposes
 *  $len>0 limits the output to $len lines
 */
int b_out(struct sbase *b,int len){
  int i;if(len==0)len=b->num;
  printf("\n sign=%s norm2=%4g num=%4d o=%d scfg=%s",
    outphase(b->sign),(double)b->norm2,b->num,b->o,ccfg(b->scfg));
  for(i=0;i<len && i<b->num;i++) 
    printf("\n%3d %s %s",i,outphase(b->is[i]),ccfg(b->ib[i]));
  return 0;
}

#if VecType & 4		/* used in b_getbase... */
#define ABS1(x) ((int)(fabs(x)+.5))
#else
#define ABS1(x) abs(x)
#endif

/* --- ax|vx>+...+1|vx> => (ax+1)|vx>+...+0|vx> --- ib[num] already stored
 * replace 0-entries by (b.num-1)-entries and b.num-- (pack) ???? 
 * the same cfg-entries with different signs do interfere to 0! 
 * twice=0: the caller function knows the twice factor already
 *     we dont have to recalculate it (ToDo: check!)
 * called by b_getbase()
 */
void b_twice(struct sbase *b,int twice){  
  int k,j,num/* ,t */;
  if(twice==0)
  for(num=0;num<b->num;num++) if (b->ib[num]==b->scfg) twice++;
  if(b->o!=2 && twice>1)
  for(num=0;num<b->num;num++) if (ABS1(b->is[num])) {
    for(k=num+1,j=twice;k<b->num && j>0;) if (b->ib[k]==b->ib[num]) {
      b->num--;j--;
      b->is[k]=b->is[b->num];		/* reduce table size */
      b->ib[k]=b->ib[b->num];
    } else k++; /* Jan03: "else" correct? new {} */
  }
  b->norm2=b->num;
  return;
}

/* ---- get_lm_factor ---- (long64: 2^64=1.845e19)
 * N=6 s=14 s*N=84 sym=12 spin-chain, J1=J2=1
 *   u=18,d=66: ns=1e18 n1= 2888 n2max=2e16 E0=-15   2s,8s (fits to 64bit)
 *   u=20,d=64: ns=1e19 n1= 4440 n2max=1e17 E0=-62         (ERR9112 k=0)
 *   u=22,d=62: ns=9e19 n1= 6500 n2max=1e18 E0=-105  3s,19s
 *   u=24,d=60: ns=2e19 n1= 9097 n2max=9e18 E0=-144  4s,26s
 *   u=26,d=58: ns=1e19 n1=12208 n2max=2e19 E0=-179  5s,35s
 *   u=42,d=42: ns=2e19 n1=17052 n2max=2e22 E0=-309 i70=169s
 * Tnorm2 is ULLong per default, but can be set to double for big-s systems
 */
Tnorm2 b_factor_lm(tbase ib1){  
  int k,b1,i1,obj[NN][4]; /* 0,1,2,3 for empty,up,down,up+down sites */
  Tnorm2 n2, factor;
  n2=1;	
  if(sym.num_lm==nn) return(n2);
  /* compute factor (n1+n2+n3+n4)!/(n1!n2!n3!n4!) =
   *   = (n1+n2+n3+n4)!/n1! * (n2+n3+n4)!/n2! * (n3+n4)!/n3!
   *  where n1=number_of_empty_sites, n2=number_of_ups, ... per LM-spin
   */
  for(k=0;k<nn;k++){ obj[k][0]=obj[k][1]=obj[k][2]=obj[k][3]=0; } 
  for(i1=0;i1<nn;i1++){  /* count types of sites n1,..,n4 */
      b1=geti(ib1,i1); obj[sym.lm[i1]][b1]++;
  }
  for(k=0;k<nn;k++)if(sym.lm[k]==k){
      factor=over(obj[k][0]+obj[k][1]+obj[k][2]+obj[k][3],obj[k][3])
            *over(obj[k][0]+obj[k][1]+obj[k][2]          ,obj[k][2])
            *over(obj[k][0]+obj[k][1]                    ,obj[k][1]);
      if( (double)(n2*factor) != n2*(double)factor ) ERR(9110);
      n2*=factor;
  }
  return(n2);
}

/* ----------------------------------------------------------------------- */
#define Err17 {static int e0=1;if(e0){ E0("b_*_lm may be buggy! check with NOS1SYM!"); e0=0;}}

/*  effect of 2-site-LM-symmetry for tU-model considering fermionic sign (Feb04)
 *      .. => +..              .3 => +3.           33 => +33
 *      .u => +u.*parity       ud => -du
 *      uu => -uu vanished     u3 => +3u*parity
 *   LM = c(+,p[i])c(+,p[j])c(-,j)c(-,i)              */
/* Feb04: rewritten
   ToDo: test S=5/2 N=6 16+14 nk=1.5e8 sym=4 n1=1091 33s until 2S=
       does not work for LM, tU u2d0 N=8 square
   oo=0: return full expanded base (all symmetric configs)
   oo=1: return smallest + sign + norm2 (dont extend base)
   oo=2: break, if sign==0 or smaller found (like ifsmallest)
   ToDo: better return was_smallest/got_smallest/0?
 */
tbase b_getbase_lm(struct sbase *b, int oo){  
  tbase ib0,ib2; int j,b1,b2,i1,i2,n2; tphase s0,s1,s2;
  CM({},Err17,Err17)	/* signum ??? */
  n2=0;	/* count same lm-scfgs */
  /* b->num must be greater than 0 (one argument or more) */
  b->sign=b->is[0]; b->scfg=b->ib[0]; /* if we have n2==1 */
  for(j=0;(ULLong)j<(ULLong)b->num;j++){
   ib0=b->ib[j]; s0=s1=b->is[j];	/* check next config */
   if(s0==0) E0("Warning: sign[j]=0");
   if (ib0< b->scfg) { b->scfg=ib0; b->sign=s0; n2=1; }
   /* ib0 = original (smallest?), ib2 = new_lm_cfg */
   for(i1=0;i1<nn-1;i1++)for(i2=i1+1;i2<nn;i2++)if(sym.lm[i1]==sym.lm[i2]) {
     b1=geti(ib0,i1); /* site 1 within LM-site */
     b2=geti(ib0,i2); /* site 2 within LM-site */
     if(b1!=b2){ /* exchange sym-sites */
       ib2=puti(puti(ib0,i1,b2),i2,b1); s2=s1;
       /* tU,tJ fermionic sign change? call (2sites) permutation function? */
#if CM(0,1,1)
       if (zi(ib0,i1)+zi(ib0,i2)!=0) { /* examples: .u u3 (fermionic transfer) */
         s2=s1*parity(ib0 & p_mask[i1][i2]); }  /* not coding independent */
       if (zizj(ib0,i1,i2)!=0){ s2=-s1; }      /* ud (fermionic exchange) */
#endif
       if(b->scfg==ib2){	/* same lm-symmetric config stored */
         if(!phase_eq(s2,b->sign)){
         b->sign=0; b->num=0; b->norm2=0; return TBASE0;}
       }
       if(!(ib0<=ib2)){ b->ib[j]=ib0=ib2; b->is[j]=s0=s2;
         if(!(b->scfg<=ib0)){
           b->scfg=ib0; b->sign=s0; n2=0; if(oo==2) return ib2; }
         i1=i2=nn; j--; /* restart for this config */ break;
       }
     }
#if CM(0,1,1)
     else {  /* b1==b2 */
       if (abs(zi(ib0,i1))>0 /* && abs(zi(ib0,i2))>0 */){
         b->sign=0; b->num=0; b->norm2=0; return TBASE0; } /* uu + -uu */
     }
#endif
   }
   /* add factor_lm new lm-symmetric configs to table *slow!* */
   if (oo==0) { Err17; } /* not supported yet, we should not need it */
  }
  /* calculation of n2 is buggy, so we reat it here */
  for(n2=j=0;(ULLong)j<(ULLong)b->num;j++){
   if (b->ib[j]==b->scfg) n2++; /* how often is scfg stored within table */
  }
  if(oo<2){	/* (n1+n2+n3+n4)!/(n1!n2!n3!n4!) */
    Tnorm2 factor=b_factor_lm(b->scfg);
    b->norm2=b->num*factor/(unsigned)n2; /* ULLong -> int ??? */
    if( (double)(b->num*factor/n2) != b->num*(double)factor/n2 ) ERR(9111);
  }
/* printf("\n dbg3: %s o=%d n2= %d norm2= %d bnum %d",ccfg(b->scfg),oo,(int)n2,(int)b->norm2,(int)b->num); */
  if(!b->norm2) b->sign=(tphase)0;
  return(b->scfg);
}

/* b = b + Tud(b), Tud is the inversion operator, b is the state 
 * called from b_getbase() for nu==nd only
 */
void b_getbase_ud(struct sbase *b){  /* for Sz!=0 wasted CPU time! */
  tbase ib0,ib1,ib2;int i,j,nj; tphase s1;
  j=nj=b->num; ib0=b->scfg; ib1=b->ib[0];
  for(i=0;i<nj;i++)/* if(b->is[i]!=0) */{
    get_sudconfig(b->ib[i],b->is[i],&ib2,&s1);	/* s|ib2>=T|ib1> */
    if(!(ib0<=ib2)){b->scfg=ib0=ib2;b->sign=s1;if(b->o==2)return; }
    if( (ib1==ib2)){
      if(!phase_eq(s1,b->is[0])){b->sign=0;b->num=0;return;}
      j=nj;break; /* because |symcfg> = Tud|symcfg> Jan00 */
    }
    b->ib[j]=ib2; b->is[j]=s1; j++;
  }
  b->norm2=b->num=j; return;
}

/* ---- long=+uudd+uudd -> short=2uudd+0uudd ----
 * faster, if symconfig and sudconfig called one after another?
 * kuerzen als extra-funktion ??? redundanter code .._lm
 * better save index for smallest? (temporarly) SORTING???
 * create cache??? or create only different configs
 * n1_sym is ca. n1/sym, d.h. ib1==ib2 | twice>1 is rare -> faster?
 *  ToDo: same as b_smallest_new()?
 * called by xval.c: modeloverlap stringorder (ToDo: remove?)
 *           spins.c: nhm_err1 storeh2 outbasis
 */
tbase b_getbase(tbase ib1,struct sbase *b){  
  tbase ib0,ib2; int j,n2,nj=sym.numall+1; tphase s1;
  b->ib[0]=b->scfg=ib0=ib1; b->is[0]=b->sign=1; n2=1;
  if(nj*((nu==nd)?2:1)>MaxSBase)F0("MaxSBase-OVL, enlarge it!?");
  for(j=1;j<nj;){ s1=1;
    get_symconfig(ib1,((unsigned)j)-1,&ib2,&s1,1);
    if(!(ib0<=ib2)){b->scfg=ib0=ib2;b->sign=s1;if(b->o==2)return ib2;}
    if( (ib1==ib2)){
      if(!phase_eq(s1,1)){b->sign=0;b->num=0;return ib2;} n2++; }
    b->ib[j]=ib2; b->is[j]=s1; /* ToDo: better call insert_sorted here */
    j++;
  }
  b->num=nj; b->norm2=nj*n2; if(n2>1) b_twice(b,n2); /* obsolete if insert_sorted */
  if (nu==nd) b_getbase_ud(b);
#if defined(CONFIG_S1SYM)    /* ToDo: does not work at the moment for tU JJ? */
  /* b_out(b,b->num); */
  if(sym.num_lm!=nn) b_getbase_lm(b,2);
  /* b_out(b,b->num); */
#endif
  return(b->scfg);
}

/* reference function for fast b_smallest3, slow, only debugging
 * ATTENTION if is[smallestcfg]=0 !!! sign of ib1
 * 1=true 0=false(smaller symcfg), -1=false(destructiv sign)
 *  ud-test only usefull if 1st spin=d!!!faster since Jan00
 */
int b_ifsmallest(tbase ib1, struct sbase *b){ 
  int old_o;
  old_o=b->o;b->o=2; b_getbase(ib1,b); b->o=old_o;
  /* printf("\ncfg=%s: ",ccfg(ib1));b_out(b,0); */
  if (!(ib1==b->scfg)) return 0;
  if (b->norm2==0 || phase_eq(b->sign,0)) return -1;
  b->sign=1; b->num=0;				/* not given here */
  return 1;
}


/*  effect of 2-site-LM-symmetry for tU-model considering fermionic sign (Feb04)
 *      .. => +..              .3 => +3.           33 => +33
 *      .u => +u.*parity       ud => -du
 *      uu => -uu vanished     u3 => +3u*parity
 *   LM = c(+,p[i])c(+,p[j])c(-,j)c(-,i)
 *  v2.21 using sym.lnext N=8 s=10/2 nud=17+10 20s(43%) => 3s(9%) PM600MHz-pg
 */
int b_ifsmallest_lm(tbase ib0){
  int i1,i2; CM({},Err17,Err17)
  for(i1=0;i1<nn-1;i1++) { i2=sym.ln[i1];
     if(i2<nn) {  /* compare member of composed spin with next member */  
      if (geti(ib0,i1)< geti(ib0,i2)) return 0;
#if CM(0,1,1) /* (+uu) + (-uu) */
      if (geti(ib0,i1)==geti(ib0,i2) && abs(zi(ib0,i1))>0) return 0;
#endif
    }
  }
  return 1;
}

/* calculate smallest lm! (enough for smallest)
   ToDo: hard code this part or examine only lm arround changed spins
      ns*(ns+1)/2 => ns, or lm*H as optimized C++ Operator
      or lm sorted Index (proof only idx-neighbours) or lm-table
 *  v2.21 using sym.lnext N=8 s=10/2 nud=17+10 7s(25%) => 4s(17%) PM600MHz-pg
 */
tbase b_smallest_lm(tbase ib0, tphase *s1){  
  int b1,b2,i1,i2;
  CM({},Err17,Err17)	/* signum ??? */
  if (s1) ;             /* dummy access, avoiding cc warnings */
  for(i1=0;i1<nn-1;i1++) { /* l1=sym.lm[i1]; */
    b1=geti(ib0,i1); if (b1<CM(2,3,4)) { 
     for(i2=sym.ln[i1];i2<nn;i2=sym.ln[i2]) /* if(l1==sym.lm[i2]) */ {
        b2=geti(ib0,i2);
        if (b1<b2) {
          ib0=puti(puti(ib0,i1,b2),i2,b1); b1=b2; /* exchange sym-sites */
#if CM(0,1,1)
          if (zi(ib0,i1)+zi(ib0,i2)!=0) /* examples: .u u3 (fermionic transfer) */ 
            *s1*=parity(ib0 & p_mask[i1][i2]);  /* not coding independent */
          if (zizj(ib0,i1,i2)!=0) *s1=-*s1;       /* ud (fermionic exchange) */
        }
        {
          if (b1==b2 && abs(zi(ib0,i1))>0){
            *s1=0; return ib0; } /* uu + -uu */
#endif
        }
      }
    }
  }
  return(ib0);
}

/* is=0 dont store or at end, is!=0 to the top, .num < nj */

/* b->o: 0=getbase     sign,norm2,scfg,num,ib[],is[] ok
 *       1=smallest    sign,norm2,scfg ok ???
 *       2=ifsmallest  if(ret=1){ret sign,scfg,norm2} ok, num=0
 */
/* ToDo: replace by b_smallest/b_getbase + b->o=2, very similar fkt.
 * what about costs? min/mean/max clocks per call? see b_smallest() 
 * b_smallest is very similar, here we have only performance adaptions
 *  1=true 0=false(smaller permutation found) -1=false(destructive sign)
 *  -1 is returned after all ret=0 tests, else we have additional
 * return 0, not -1, for cfgs like (3.3333)
 *   this reduces compute time in new_numsymconf(v2.27)
 *   . means empty, 3 means up+down for the tU-model
 * called by ini_norm2 ns_thread
 * on S1SYM ifsmallest may fail, if other syms mix S1 subsites
 *   ToDo: Permutations must not mix sites (its bitweight) within S1 groups?
 *         do it for all commuting perms!? -> example
 *   |1234> P1=(1-3 2-4) P2=(1-2) P3=(3-4) P4=P1*P3=(1-4 2-3)=not_allowed
 * ToDo: store biggest (max) and/or smallest (min) symcfg???
 */
int b_ifsmallest3(tbase ib1, struct sbase *b){ 
  t_sites k1[NN],k2[NN],*k3,i2; int i,j,n2,d;  tphase s1,s2,s3;
  /* move k1,k2,ks to sym.buff for L1-caching? NN*MaxSym=40*410 */
  tbase ib2=ib1;
  b->sign=1; n2=1; s2=+1;
  if (nu==nd && sym.wud) get_sudconfig(ib1,+1,&ib2,&s2); /* s|ib2>=T|ib1> */
#if defined(CONFIG_S1SYM) /* return b_ifsmallest(ib1,b); */
  if (sym.num_lm!=nn) {
    if (!b_ifsmallest_lm(ib1)) return 0; /* never hits!? tU? */
    if (nu==nd && sym.wud) {
       ib2=b_smallest_lm(ib2,&s2); /* smallest inverse cfg */
       if (s2==0) { /* destructive sign, tU+tJ only, never hits? */
         ERR(9999); return -1;
       }
    }
  }
#endif
  if (nu==nd && sym.wud) {
    if (!(ib1<=ib2)) return 0; /* inverse config is smaller */
    if ( (ib1==ib2)) { // would also be done in the permutation loop?
      if (s2!=1) { b->sign=0; } /* destructive phase, tU only? (.333) */
      n2=2;
    }
    for(i=0;i<nn;i++) k2[i]=geti(ib2,i);
  }
  for(i=0;i<nn;i++) k1[i]=geti(ib1,i);  /* 30% CPU */
  for(i2=0,k3=k1,s3=1;;i2++,k3=k2,s3=s2){  /* ud=+1,-1 */
    /* search for smallest symmetric config as representator */
    for(j=0;j<sym.numall;j++){
      i=nn; do { i--;	/* Tj, high -> low */
        /* backward loop can be replaced by forward loop easily */
        d=k1[i]-k3[ sym.i[j][i] ];  /* permutation, 40% CPU */
      } while (d==0 && i!=0);       /* permute as long as bits are equal */
      if (d<0) continue;            /* permutation is greater */
      if (d!=0) { return 0; }       /* permutation is smaller */
#if CM(1,0,0)  /* signum is not cfg-dependend */
      s1=sym.w[j];                  /* T|x>=t|x> same config */
#elif CM(0,1,1)  /* signum is cfg-dependend */
      tbase ib4; s1=1;
      get_symconfig(((i2)?ib2:ib1),j,&ib4,&s1,1);
#endif
      if(!phase_eq(s1*s3,1)) b->sign=0; /* return (3.3333)=0 not -1 */
      n2++;
    }
    if(nu!=nd || i2>=1 || !sym.wud) break; /* Sep00 */
    IFww( break; )
  }
  b->scfg=ib1; b->num=0; /* not given here */
  b->norm2=(((nu==nd && sym.wud)?2:1)*(sym.numall+1))/n2; /* OK ??? */
#ifdef CONFIG_S1SYM
  { Tnorm2 factor=b_factor_lm(ib1); /* check for 64bit overflow */
    if( (double)(b->norm2*factor) != b->norm2*(double)factor ) ERR(9112);
    b->norm2*=factor;	/* Ok ??? */
  }
#endif
  if (b->sign==0) return -1; /* Oct07 2.27 */
  /* b_out(b,0); */
  return 1;
}


/* --- last change 1998-169 JS --- */
/* static inline */ /* needed in spins.h + xval.c */
/*
   spins.h struct sbase, tbase scfg, Tnorm2 norm2, tphase sign, int num
   work variables:
    k1 - config ib1
    k2 - config ib1 or spin-inversed ib1 (not used for nu!=nd)
    ks - smallest config found actually 
   used by spins.c nhm() 107/135s, use b->sign, b->norm2 and b->scfg
   ToDo: remove b->bi[] array (for getbase call remalloc)
   see b_smallest_old (old version 10times faster)
   called by: nothing (simpler code for FPGA replacement?)
 */
tbase b_smallest_new(tbase ib1, struct sbase *b){ /* about 63% CPU -O1 10+10 */
  int j,i2,n2; tphase s1,s2,s3;
  tbase ib2, ib3, ib4;
  b->sign=1;  b->scfg=ib1; b->norm2=1;	/* default minimum */
  n2=1; get_sudconfig(ib1,+1,&ib2,&s2);	/* s|ib2>=T|ib1> */
#if CM(0,1,1)   /* v2.0.2++ ToDo: get fast tJ8u1d2e1 running ??? */
  { b->o=1;b_getbase(ib1,b);b->o=0; return(b->scfg); } /* wont work for LM */
#endif
#if defined(CONFIG_S1SYM)
  if(sym.num_lm!=nn){ 
                      b->scfg=ib1=b_smallest_lm(ib1,&(b->sign));
                      if (nu==nd) ib2=b_smallest_lm(ib2,&s2);
                      if (s2==0 || b->sign==0){ b->sign=0; return TBASE0; }
                      b->norm2*=b_factor_lm(ib1); /* fermionic sign? */
                    }
#endif
  if (nu==nd && sym.wud){
    if((ib2==ib1)){ /* better do it later, because its rarely */
     n2=2;
     if(s2!=1){ b->sign=0; return TBASE0;}	/* s1!=1 */
    }
    if(!(ib1<=ib2)){ 
     ib3=ib1;b->scfg=ib1=ib2;ib2=ib3; b->sign=s2;s2=1; }	/* new minimum */
  }
  for(i2=0,s3=b->sign,ib3=ib1; ; i2++,s3=s2,ib3=ib2){  /* ud=+1,-1 */
    /* search for minimum symmetric config as representator */
    for(j=0;j<sym.numall;j++){		/* take a SymOp=PermOp(j)=ps */
      /* better call/inline get_symconfig ? */
      get_symconfig(ib3,j,&ib4,&s1,1);
      if (ib4> b->scfg) continue; /* its greater, forget it (how often?) */
      if (ib4==b->scfg) {    /* same (how often?) phase -> int */
        if(!phase_eq(s1*s3,b->sign)){b->sign=0;return TBASE0;}
        n2++; continue;
      }
      b->scfg=ib4; /* found smaller cfg */
      n2=1;
      b->sign=s1*s3; /* better use k(2PIi/l) */
    }   /* test next sym */
    if(nu!=nd || i2>=1 || !sym.wud) break; /* Sep00 */
    IFww( break );
  }     /* test ud-sym */
  b->num=0;				/* not given here */
  b->norm2*=(((nu==nd && sym.wud)?2:1)*(sym.numall+1))/n2; /* Ok? */
  /* if(norm2) b->norm2=(unsigned Tnorm2)norm2[b2i(b->scfg)]; */
  return b->scfg;
}

/* called by b2i_err nhm_err1 nhm hamilton_nhv? wopij_block
 *  ToDo: array of ib1's, output array, return number if calculation stopped?
 *  ToDo: better return 0 or 1=smallest to replace if_smallest
 *        input ib1 via *b (less arguments, type tbase not need to declared)
 *   10times faster than old code calling get_symconfig numsym times
 * called by: xval.c.mrule ...
 */
tbase b_smallest(tbase ib1, struct sbase *b){ /* about 63% CPU -O1 10+10 */
  t_sites k1[NN],k2[NN],ks[NN],*k3; 
  /* move k1,k2,ks to sym.buff for L1-caching? NN*MaxSym=40*410 */
  int i,j,i2,n2,d; tphase s1,s2,s3;
  tbase ib2,ib3;
  /* set default minimum, better set by caller? also reduces arguments */
  b->sign=1;  b->scfg=ib1; b->norm2=1;
  n2=1; get_sudconfig(ib1,+1,&ib2,&s2);	/* s|ib2>=T|ib1> */
#if CM(0,1,1)   /* v2.0.2++ ToDo: get fast tJ8u1d2e1 running ??? */
  { b->o=1;b_getbase(ib1,b);b->o=0; return(b->scfg); } /* wont work for LM */
#endif
#if defined(CONFIG_S1SYM)
  if(sym.num_lm!=nn){ 
                      b->scfg=ib1=b_smallest_lm(ib1,&(b->sign));
                      if (nu==nd) ib2=b_smallest_lm(ib2,&s2);
                      if (s2==0 || b->sign==0){ b->sign=0; return TBASE0; }
                      b->norm2*=b_factor_lm(ib1); /* fermionic sign? */
                    }
#endif
  if (nu==nd && sym.wud){
    if((ib2==ib1)){ /* better do it later, because its rarely */
     n2=2;
     if(s2!=1){ b->sign=0; return TBASE0;}	/* s1!=1 */
    }
    if(!(ib1<=ib2)){ 
     ib3=ib1;b->scfg=ib1=ib2;ib2=ib3; b->sign=s2;s2=1; }	/* new minimum */
  }
  for(i=0;i<nn;i++){ ks[i]=k1[i]=geti(ib1,i); k2[i]=geti(ib2,i); }
  for(i2=0,k3=k1,s3=b->sign,ib3=ib1;;i2++,k3=k2,s3=s2,ib3=ib2){  /* ud=+1,-1 */
    /* search for smallest symmetric config as representator */
    for(j=0;j<sym.numall;j++){		/* take a SymOp(j)=Permutation */
      /* better call/inline get_symconfig ? */
      i=nn; do { i--;                 /* Tj, high -> low */
        /* cache miss? -> no, its accessed ca. 700 times per call, 14s */
        /* 32+8: J2=.5 0.10205747 0.25766905 53s,5.5m -> 49s,4.9m */
        /* ToDo: check only occupied sites? or use hash tables */
        /* try forward-access per sym.iback[j][i] too */
        /* 32+8: 700*hnz=4.4*sym*hnz max=N*sym*hnz */
        d=ks[i]-k3[ sym.i[j][i] ];  /* compare cfg with SymOp*cfg */
        /* [j][i]-access vs. pointer good for x86, bad for sparc? */
        /* check if cfgs are still equal after permutation of site i */
      } while (d==0 && i!=0); /* true: 34+6=83% 32+8=78% */
      if (d<0) continue; /* greater cfg, 157*hnz calls, true(32+8)=99.2% */
#if     CM(1,0,0)  /* signum is not cfg-dependend */
      s1=sym.w[ j ]; 
#elif   CM(0,1,1)  /* signum is cfg-dependend */
      tbase ib4; get_symconfig(ib3,j,&ib4,&s1,1); /* permute sites */
#endif
      if (d!=0) {  /* 32+8 J2=.5: 1.2*hnz calls, true: 99.7% smaller cfg */
        for(;i>=0;i--) ks[i]=k3[ sym.i[j][i] ]; /* store rest config */
        n2=1; b->sign=s1*s3; continue; /* found smaller cfg */
      }
      /* get same config: S|k> to s|k> */
      /* sign=0? 32+8 true=0..3% sym_k-depending, better use k(2PIi/l)? */
      if(!phase_eq(s1*s3,b->sign)){b->sign=0;return TBASE0;}
      n2++;
    }   /* test next sym */
    if(nu!=nd || i2>=1 || !sym.wud) break; /* Sep00 */
    IFww( break );
  }     /* test ud-sym */
  encode_ud( &(b->scfg), ks, nn); /* call operator already with decoded cfg? */
  b->num=0;				/* not given here */
  b->norm2*=(((nu==nd && sym.wud)?2:1)*(sym.numall+1))/n2; /* Ok? */
  /* if(norm2) b->norm2=(unsigned Tnorm2)norm2[b2i(b->scfg)]; */
  return b->scfg;
}


#define BLOCKS	8	/* can be only 2^n=1,2,4,8,etc */
/* version 0.1 Feb 2000 first try version, not finaly, for checks 
 * ideal way for part l1 into blocks => multi processing
 * in ideal case fast and invariant against symmetry operations 
 * min=0 max=n*(n-1)/2 ??? besser mit Untergittern
 */
int block_no(tbase ib1){ /* use pair correlations, count d-d pairs */
  int no,i1,i2; no=0;
  for(i1=0;i1<nn;i1++) if (zi(ib1,i1)==-1) /* nn+nd*nn Ops */
  for(i2=0;i2<nn;i2++) if (zi(ib1,i2)==+1)
   no+=sym.corr[i2][i1];
  return (no>>1)&(BLOCKS-1);
} 

/* --- all what be needed for pthreads of numsonf --- */
typedef struct sp_ns {
#ifdef CONFIG_PTHREAD
  pthread_t pt;
#endif
  int block;        /* block nr (0..threads*nodes-1) for debugging */
  ULLong i1,n;      /* idx range of configuration to check (r) */
  ULLong numans;    /* num_ansatz (w) */
  Tnorm2 norm2;     /* check sum (w) */
  ULLong num;       /* number of stored sym configs (w) */
  ULLong num0;      /* number of sym configs with sign=0 (w) v2.27 */
  tbase ib1;        /* spin-configuration (t) */
  struct sbase b;   /* working buffer (t,big) */
  struct sp_ns *next; /* pointer to next data set */
  char fname[32];     /* filename tmp/l1_%3d.tmp (r) Oct07 */
} tp_ns;

/* void-function for pthreads called by numsymconf(), see below */
#define L1_BUFLEN (8*1024*1024)  /* in units of Bsize, ca. 40MB for 1-10write/s */
void ns_thread( void *ptr ){
    tp_ns *p=(tp_ns *)ptr;
    tbase ibold; 
    ULLong i1, i2, i3; long t1=itime();
    FILE *fl1=NULL;
    unsigned char *buf;
    ULLong i_out=1l<<14; /* counter for output, speed estimation */
    int rt;              /* return code from b_ifsmallest3 */
    fl1=fopen(p->fname,"wb"); if(!fl1) F1("fopen_%s",p->fname); i2=0;
    buf=(unsigned char *)malloc(L1_BUFLEN*Bsize); if (!buf) F1("malloc node %d",mpi_i);
    for( ibold = p->ib1 = ll1(i1=p->i1); i1 < p->n; ){
#ifdef CONFIG_ANSATZ
      if (!ansatz_test(p->ib1)) p->numans++; else
#endif
      {
        rt=b_ifsmallest3(p->ib1,&(p->b)); /* 1=true, 0=false -1=sign0 */
        if (1==rt) { /* 1=true, 0,-1=false */
          if (i2 >= L1_BUFLEN) {  /* write out full buffer */
            i3=fwrite(buf,Bsize,i2,fl1); if(i3!=i2)F1("fwrite %s",p->fname);
            i2=0;
          }
          putl1(i2,p->ib1,buf);   /* store mincfg to l1p[l1s] */
          p->num++; i2++;         /* count stored cfgs */
          p->norm2+=p->b.norm2;
        }
        if (-1==rt) { p->num0++; p->norm2+=p->b.norm2; }
      }
      i1++;                     /* loop-counter */
      p->ib1=next(p->ib1);      /* next higher config */
      if (p->ib1<=ibold) break; /* avoid overrun */
      ibold=p->ib1;
      if (i1==i_out && mpi_i==0 && p->block==0) { long t2=itime();
        i_out<<=2; /* x4 */
        if (t2-t1>8)
        printf("\n ns %6.2f%% found= %9.0f of %9.0f t= %6.2fm t100%%= %6.2fm",
               i1*100./p->n, 1.*p->num, 1.*i1, (t2-t1)/60., (t2-t1)/60.*p->n/i1 );
      }
    }
    /* ToDo: output only min/mean/max in the main thread later */
    if (mpi_i*pt_n+p->block%pt_n<8) /* ToDo: output first and last four */
      printf("\n ns.thread %2d %2d from= %9.0f len= %9.0f found= %8.0f t=%s",
        mpi_i, p->block, (double)(p->i1), (double)(p->n - p->i1), (double)p->num,
        sclock(itime()-t1)); /* block=0..mpi_n*B_NUM-1 */
    i3=fwrite(buf,Bsize,i2,fl1); if(i3!=i2)F1("fwrite %s",p->fname);
    fclose(fl1); /* flush */
    free(buf);
}


/* number of (symmetric) base-states(NN,nu,nd) -> n1_using_Sym 
 *  global vars: sym,nu,nd!
 * until v2.19:
 *   dividing in equal blocks for threads is unbalanced (2/3 only one cpu)
 *   because 1st thread has more work as the others (3 times more)
 * ToDo: better thread-balance
 *   1) create nt=4 threads, set results=0
 *      pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
 *      better pthread_mutex_init(&mutex1,NULL)
 *   2) pthread_mutex_lock (only one thread can pass)  (OMP_critical)
 *   3) store results, reset results, get new work=block(ns/(16*nt)) (serialized)
 *   4) pthread_mutex_unlock
 *   5) does the job and return to (2) -or- leave if work=0
 *   ... (MPI_Barrier, OMP_barrier)
 *   6) destroy threads
 * exsample: jj square 160sym nn=36+4 sym_k=0 ns(91390)=632=523+85+22+1+1
 * ToDo: try IsingGS + non-sym-excitations until IsingMax (how to divide?)
 * ToDo: dont use disk (MPI+NFS is slow and instable)
 *    better use MPI-disk emulation OR parallel CPU and save local part
 *    store buffered blocks to node (i++%mpi_n)?
 * ToDo: estimation of n1? upper boundary?
 *        MC? 100*blocks random symmetrized startvec
 *            divided into blocks + storeh2?
 *            hash = ising+meanerg_of_excitations
 *  (minimize off_diag_block_elements for maximum performance)
 *   
 */
ULLong numsymconf(){  /* short: ns() */
  ULLong i2=0, i3, numans=0, num0=0;
  Tnorm2 sum=0;
  tp_ns pns, *ppns, *p2;
  int ii, i4;
extern int algo;
  p2=NULL ; /* dummy access, avoiding cc warnings */
  ppns=&pns; i3=numconf(); i4=pt_n*mpi_n; /* i3=(N!)(nu!nd!) i4=num_threads */
  /* the estimation of n1 failes if there are local symmetries. n1 bigger? */
  if (!mpi_i)
    printf(" ncfgs=%.2e nsym=%d estimated nscfgs n1: %.2e\n",  
             i3*1., sym.numall+1, i3*1./(sym.numall+1));
  if (vvv && !mpi_i) printf(" numsymconf(%.15g)/(mpi*pt=%d)", (double)i3, i4);
      MPI_Barrier(MPI_COMM_WORLD);

  if(!(algo&16) && !mpi_i)  /* new method v2.27, old method else */
  { /* ToDo: tJ + tU, nd>nu is much faster, test=algo&16 */
    /* number of b_ifsmallest calls dominates compute time
     *        so we have a speedup for sym>=2
     * time: n1=n!/(nu!nd!)/sym, old=n1/600000[min] new=sum[i=0..nu] n1s(i)/96e6
     * (40!/8!/32!)=76904685=480654*160+45=482184+482139+480948+...
     * JJ worst case: numsym=0 -> 1..10% slower
     * sym_k != 0 works, ToDo: tU,tJ
     * ToDo: if nu>nd do inversion! tU -> .3 inversion + U -> -U
     *                                for U*(niu-1/2)*(nid-1/2)==Ham2
     *        for nu>nd do inversion before nsymconf and again before storing
     *
     *  see ../doc/numsymcfg.html (explanation+speed+test+examples)
     *   tJ code checked, tU checked (tune+further checks later!)
     */
    extern ULong x_max;
    struct sbase b;   /* working buffer (big) */
    int tnu=nu, tnd=nd; /* store the original nu,nd and restore later */
    tbase state[2*NN+1]; /* rekursion buffer (max nu+nd) */
    int id[2*NN+1], iu[2*NN+1]; /* number of bits set */
    int rd=0; /* rekursion depth */
    int rs;   /* return ifsmallest */
    int hus[2*NN+1]; /* highest unset site(bit), goes from high to low */
    int nbp[2*NN+1]; /* next bitpattern, 3=u+d 2=d 1=u 0=. */
    unsigned char *buf; ULong bi2; /* write buffer and its counter */
    FILE *fl1=NULL;
    tp_ns *p=&pns;
    long t2, t1=itime();
    ULLong i_out=1L<<8; /* output of estimated time */
    ULLong nt=0l;        /* number tested cfgs for speed analyse */
    if (!mpi_i) printf("\n !(algo&16): new ns, 10xfast+single_thread ");
    if (vvv && !mpi_i)
      printf("\n new_numsymconf(%.15g)/(mpi*pt=%d) nd>nu,"
       " nQ>n. is (1-8x) faster", 1.*i3, i4);
    p->norm2=0;   p->numans=0; p->b.o=2; p->num=0;
    sprintf(p->fname,"tmp/l1_%03d.dat",0);
    fl1=fopen(p->fname,"wb"); if(!fl1) F1("fopen_%s",p->fname); bi2=0;
    buf=(unsigned char *)malloc(L1_BUFLEN*Bsize); if (!buf) F1("malloc node %d",mpi_i);
    b.o=2; /* ??? */
    /* the initial state is all sites with maximum bits set (biggest) */
    state[0]=TBASE0;
    for(ii=0;ii<nn;ii++) state[0]=puti(state[0],ii,CM(2,2,3));
    hus[0]=nn; nbp[0]=CM(0,-1,-1); id[0]=nn; iu[0]=CM(0,0,nn);
    i2=0; num0=0; sum=0; /* counter scfg states + scfgs_with_sign0 */
    if (vvv&4)
      printf("\n start      %s rd= %2d hus= %2d",
             ccfg(state[rd]), rd, hus[rd]);
    /* now reduce bits from highest bits, test if smallest and rekursiv */
    /* ToDo: hus and nn-rd is the same! remove hus[] */
    for(;;) {
      nbp[rd]++;
      if (nbp[rd]>CM(1,1,2)) {
        hus[rd]--; /* down to next lower site */
        nbp[rd]=CM(1,0,0);
      }
      if (hus[rd]<=0 CM(|| tnu-iu[rd]>hus[rd],\
                        || id[rd]-hus[rd]>tnd || id[rd]<tnd || iu[rd]>tnu || iu[rd]+hus[rd]<tnu || id[rd]+iu[rd]<tnu+tnd,\
                        || hus[rd]<iu[rd]-tnu || hus[rd]<id[rd]-tnd)) {
        /* now we have tried all sites, so we can return */
        if (rd==0) break; /* all work done */
        rd--;     /* go up */
//        if (i2<x_max && vvv&4)
//          printf("\n come back  %s rd= %2d hus= %2d %2d+%2d %.0f",
//                 ccfg(state[rd]), rd, hus[rd], iu[rd], id[rd],1.*nt);
        continue; /* return and try next deeper state */
      }
      /* probewise reset bit at hus */
      hus[rd+1]=hus[rd]-1;/* go to next lower site index when going deeper */
      nbp[rd+1]=CM(0,-1,-1);
      /* set nu, nd to have correct working behaviour of b_smallest */
      id[rd+1]=id[rd]-1+(nbp[rd]>>1); /* 1: nd--  0: nd-- */
      iu[rd+1]=iu[rd]+((1&nbp[rd])?CM(1,1,0):CM(0,0,-1));  /* 1: nu++  0: nu-- */
      nu=0; nd=nn; /* trick to avoid sym_ud to early */
      CM(,if (iu[rd+1]!=tnu && id[rd+1]==tnd) { continue; },)
      CM(,,if (iu[rd+1]<tnu || id[rd+1]<tnd) { continue; })
      /* much faster for N=40 1u1d */
      CM(,,if (iu[rd+1]-hus[rd+1]>tnu || id[rd+1]-hus[rd+1]>tnd) { continue; })
      CM(,if(id[rd+1]<tnd || iu[rd+1]>tnu || iu[rd+1]+hus[rd+1]<tnu || id[rd+1]+iu[rd+1]<tnu+tnd) continue;,)
      state[rd+1]=puti(state[rd],hus[rd+1],nbp[rd]); /* lower bit pattern, down -> up */
      /* go deeper if we have only one possibility to set the rest cfg */
      CM(,,\
      if (hus[rd+1]>0 && id[rd+1]-tnd==hus[rd+1] \
                      && iu[rd+1]-tnu==hus[rd+1]) { rd++; continue; })
      CM(,,\
      if (hus[rd+1]>0 && id[rd+1]-tnd==hus[rd+1] \
                      && iu[rd+1]-tnu==0) { rd++; continue; })
      CM(,,\
      if (hus[rd+1]>0 && iu[rd+1]-tnu==hus[rd+1] \
                      && id[rd+1]-tnd==0) { rd++; continue; })
      CM(,,\
      if (hus[rd+1]>0 && iu[rd+1]-tnu==hus[rd+1] \
                      && id[rd+1]    ==hus[rd+1]) { rd++; continue; })
#if 0
      CM(,,\
      if (hus[rd+1]>0 && iu[rd+1]    ==hus[rd+1] \
                      && id[rd+1]-tnd==hus[rd+1]) { rd++; continue; })
#endif
      // 38+1
      CM(,\
      if (hus[rd+1]>0 && id[rd+1]-hus[rd+1]==tnd \
                      ) { rd++; continue; },)
#if 0
      // 38+1
      CM(,\
      if (hus[rd+1]>tnd && id[rd+1]==hus[rd+1] \
                      ) { rd++; continue; },)
#endif
      // 39+1
      CM(\
      if (hus[rd+1]>0 && iu[rd+1]-tnu==hus[rd+1] \
                      ) { rd++; continue; },,)
      CM(\
      if (hus[rd+1]>0 && id[rd+1]-tnd==hus[rd+1] \
                      ) { rd++; continue; },,)

      if (iu[rd+1]==tnu && id[rd+1]==tnd) {nu=tnu; nd=tnd;} /* sym.ud=on */
      rs=b_ifsmallest3(state[rd+1],&b); /* 1=true 0=false -1=destr_sign */
      nt++; /* count ifsmallest calls for speed output */
      if (i2<x_max && (vvv&5)==5) // || hus[rd+1]!=nn-rd-1)
      printf("\n try  % d    %s rd= %2d hus= %2d %2d+%2d",
        rs, ccfg(state[rd+1]),rd+1, hus[rd+1],iu[rd+1],id[rd+1]);
      if (rs) { /* go deeper or count for 1=true and -1=destructiv_sign */
        if (iu[rd+1]==tnu && id[rd+1]==tnd) { /* found symmmetric state */
#ifdef CONFIG_ANSATZ
          if (!ansatz_test(p->ib1)) p->numans++; else
#endif
          {
            if (rs==1) {  /* without destructive phase */
              i2++;
              {
                if (bi2 >= L1_BUFLEN) {  /* write out full buffer */
                  if (fwrite(buf,Bsize,bi2,fl1)!=bi2) F1("fwrite %s",p->fname);
                  bi2=0;
                }
                putl1(bi2,state[rd+1],buf);   /* store mincfg to l1p[l1s] */
                p->num++; bi2++;         /* count stored cfgs */
                p->norm2+=b.norm2;
              }
              if (rd+1>2*NN || i2<x_max) //  && vvv)
                printf("\n found      %s rd= %2d hus= %2d %2d+%2d %4.0f",
                      ccfg(state[rd+1]),rd,hus[rd+1],iu[rd+1],id[rd+1],1.*i2);
              /* found: hus[rd+1]=nn-(rd+2) */
              /* ToDo: output time estimation if i2 has 4folded */
              if (i2==i_out) {
                t2=itime();
                if (t2-t1>8)
                printf("\n ns %6.2f%% found= %9.0f of %9.0f t= %6.2fm t100%%= %6.2fm",
                   i2*100./(i3*1./(sym.numall+1)), 1.*i2, 1.*nt, (t2-t1)/60., (t2-t1)/60.*((i3*1./(sym.numall+1))/i2) );
                i_out<<=2; /* x4 */
              }
            } /* rs==1 found smallest, nondestructive sign */
            else if (rs==-1) { num0++; p->norm2+=b.norm2; } /* count sign0 */
          }
        } else {
          /* found symmetric mother cfg, go deeper */
          rd++; /* go deeper, ToDo: start by putting rest up's to next sites? */
          if (i2<x_max && vvv&4)
            printf("\n dive into  %s rd= %2d hus= %2d %2d+%2d %.0f",
                   ccfg(state[rd]), rd, hus[rd], iu[rd], id[rd],1.*nt);
            /* ToDo: store found+nt, on return print found as success rate */
          if (rd>2*NN) { ERR(9999); } continue;
        }
      } /* smallest cfg */
//      else { nbp[rd]+=2; } /* 38+1 but 37+1? if . not possible, u also dont work??? tJ++ */
    } /* for(;;) */
    sum=p->norm2;
    nu=tnu; nd=tnd; /* restore gobal nu, nd */
    if (fwrite(buf,Bsize,bi2,fl1)!=bi2) F1("fwrite %s",p->fname);
    fclose(fl1); /* flush */
    free(buf);
    printf("\n numscfg2.27 scalls= %8.0f found= %8lu sgn0= %6lu ns.t= %.2f min",
           1.*nt, (ULong)i2, (ULong)num0, (itime()-t1)/60.);
  }
#ifdef CONFIG_MPI
  if (!(algo&16)) {
    /* send (i2) from root and receive (i2) by all other ranks */
    MPI_Bcast(&i2,sizeof(ULLong),MPI_BYTE,0,MPI_COMM_WORLD);
  }
#endif

  if (algo&16) /* old pre2.27 method, obsolete */
  {
    if (!mpi_i) printf("\n try !(algo&16) to use the new ns-algo, 10xfaster");
#if CM(1,0,0)
    if (nu==nd && sym.num_lm==nn && sym.wud) i3/=2;  /* ??? does not work for LM<NN */
#endif                  
#ifdef CONFIG_S1SYM
    if (sym.num_lm!=nn) { /* avoid a bug for (s=1,JJ,mpt), cfgs counted twice */
      i4=1; /* ToDo: try to remove this again, bad sum_n2! */
      printf("\n program switches temporarely to single thread, avoiding a bug (s=1,JJ ToDo)");
    }
#endif
    numans=i2=0; num0=0; sum=0;
    /* write to files within threads to avoid unbalanced memory */
    for ( ii=mpi_i*pt_n; ii<i4 && ii<(mpi_i+1)*pt_n; ii++ ) {
      if (!ppns) ERR(9011);                     /* check buffer for thread */
      ppns->block=ii;                           /* block number */
      /* v2.26 (propto 2^ii)=unbalanced, try ii^2?, v2.27 (propto ii) */
      ppns->i1=(i3+i4-1)/i4*(ii);               /* start of block */
      ppns->n =(i3+i4-1)/i4*(ii+1);             /* start of next block */
      if (ppns->i1>i3) ppns->i1=i3;             /* aware of rest block */
      if (ppns->n >i3) ppns->n =i3;             /* aware of rest block */
      ppns->norm2=0;   ppns->numans=0; ppns->b.o=2; ppns->num=0; ppns->num0=0;
      sprintf(ppns->fname,"tmp/l1_%03d.tmp",ii);
      /* each thread writes its file, aware of max file handles! */
      parallel_start(ii%pt_n, ns_thread, ppns);
      if( ii<i4-1 ) ppns->next = (tp_ns *) malloc(sizeof(tp_ns));
      ppns=ppns->next;
    }
    MPI_Barrier(MPI_COMM_WORLD); /* debug */
    if (vvv && !mpi_i) printf("\n join threads");   /* easier to debug */
    for ( ppns=&pns,ii=mpi_i*pt_n; ii<i4 && ii<(mpi_i+1)*pt_n; ii++ ) {
      parallel_wait(ii%pt_n);
      // save_l1(ppns->l1p,ppns->l1s,i2); /* write to tmp/l1_%03d.dat @pos i2 */
         /* uchar *l1p, ulong stored, ulong ofs */
      i2+=ppns->num; sum+=ppns->norm2; numans+=ppns->numans; num0+=ppns->num0;
      if (vvv&2 && (ii<4 || ii>=mpi_n*pt_n-4))
      printf("\n node %2d n1= %.15g n_ans= %.15g sum_n2= %.15g t+= %s",
             mpi_i, (double)i2, (double)numans, (double)sum, sclock(~0l));
      p2=ppns; ppns=ppns->next; if (ii>0) free(p2); /* free after access! */
    }
#ifdef CONFIG_MPI
    {
/* on Tru64+SLES9 MPI_ULLong is not available */
#if (!defined(MPI_UNSIGNED_LONG_LONG)) && defined(MPI_LONG_LONG_INT)
#define MPI_UNSIGNED_LONG_LONG MPI_LONG_LONG_INT
#endif
      // MPI_UNSIGNED_LONG_LONG 
      ULLong lsend[4], lrecv[4];
      int err; 
      lsend[0]=i2;
      lsend[1]=numans;
      lsend[2]=sum;
      lsend[3]=num0;
      err=MPI_Allreduce( lsend, lrecv, 4,
            ((sizeof(ULLong)==8)?MPI_UNSIGNED_LONG_LONG:MPI_UNSIGNED_LONG),
                         MPI_SUM, MPI_COMM_WORLD);
      i2=    lrecv[0];
      numans=lrecv[1];
      sum=   lrecv[2];
      num0=  lrecv[3];
      /* min/max time per thread and node */
    }
#endif
    /* ToDo: concat cat l1_*.tmp > l1_*.dat # replace by better function */
    if (!mpi_i) { int ret;
                  ret=system("cat tmp/l1_*.tmp > tmp/l1_000.dat");
                  if (ret) F1(" system() err=%d",ret);
    }
    MPI_Barrier(MPI_COMM_WORLD); /* wait for node0 */
  }

  if (!mpi_i) {
    printf("\n ns_end: n1= %.15g sgn0= %.8g n_ans= %.15g sum_n2= %.15g t+= %s",
           (double)i2, (double)num0, (double)numans, (double)sum, sclock(~0l));
    if (sum!=i3 && !mpi_i) printf(" sum!=i3");
  }
  return(i2);
}

/* return a unique(?) list of counted different pairs = ising energy index
   used to block hilbertspace using model-symmetries in future versions
   uu+dd ud+du ..+33 .3+3. .u+.d+u.+d.+3u+3d+u3+d3  (16pairs -> 5sympairs)
   the list is allocated and must be freed after use
   mode|=2: output list
   mode|=1: return list (else free list, return 0)
   buf = result-array or NULL (if NULL, alloc memory)
   called by: ini_norm2 storeh2 x_out => replace by
      int get_ising_erg(b1,mode,weights)
      return product j1ud*a1+j2ud*a2+... removes need of storing array
 */
int *get_dcorr(tbase b1, int mode, int *buf) {
  int *result, i1, i2, i3, ib1, ib2;
  /* map an site pair nisnjs to a Ising index to count up */
  const char map[16]= /* fits to L1 cache line !? */
    {  2, 3, 3, 4 ,  3, 0, 1, 3 ,  3, 1, 0, 3 ,  4, 3, 3, 2 };
   /* .. .u .d .3   u. uu ud u3   d. du dd d3   3. 3u 3d 33 => 0..4 = 3bit */
   /* index: 0=uu+dd  1=ud+du  [2=..+33  3=.u+u.+.d+d.  [4=.3+3.]] */
   if (buf) result = buf;
       else result = (int *) malloc(sym.num_dcorr*CM(2,4,5)*sizeof(int));
   if (result) {
     for (i1=0;i1<sym.num_dcorr*CM(2,4,5);i1++) { result[i1]=0; } /* reset */
     for (i2=   0;i2<nn-1;i2++)    /* N*(N-1) site pairs */
     for (i3=i2+1;i3<nn  ;i3++) {
           i1=sym.corr[i2][i3];    /* access N*N pair array */
           ib1=geti(b1,i2);  /* 0=. 1=u=o 2=d=x 3=u+d=Q */
           ib2=geti(b1,i3);
           // printf("DBG: %d %d  %d %d  %d\n",i2,i3,ib1,ib2,i1);
           /* PM600MHz 32+8 isinghash=14s for 4*if(&&||)else */
           /* PM600MHz 32+8 isinghash=17s(int),12s(char) for array-access */
           result[CM(2,4,5)*i1+map[4*ib1+ib2]]++;
           // => ising+=weight[CM(2,4,5)*i1+map[4*ib1+ib2]]; /* +Ji1 for ud */
     }
     if (mode&2) {
       /* printf("\n dcorr: cfg=%s", ccfg(b1)); */
       printf(" dcorr:");  /* different pair correlations */
       for (i1=1;i1<sym.num_dcorr;i1++)
        for (i2=0; i2<CM(2,4,5)-1; i2++) /* dont print CM(uu,.u,.3) */
         printf(" %2d",result[CM(2,4,5)*i1+i2]); /* [j1,ud-.u] [j2,ud-.u] .. */
     }
   }
   /* free if temporary allocated, for output only */
   if ((!(mode&1)) && (!buf) && result) { free(result); result=NULL; }
   return result;
}

/* idea: sort l1 for norm2 (better ising-erg), to avoid storing norm2 */
#define N2Blen 1021  /* histogram size, prime number */
int     n2b[N2Blen]; /* histogram, hashed */
Tnorm2  n2h[N2Blen]; /* hash table, norm2 */

tbase *b_l1=NULL; /* table: first_sym-config of each block (global) */

extern ULong x_max; /* ToDo: maximum number output xval.h => debug.h */

/* ----- initialize symmetric base (cfg-tables l1) v1.53 ----- 
 * calculate orbit length or norm of each sym-config
 *  also do some speed tests and test new ideas
 * ToDo: should be done with creation of l1
 * v2.33 norm2[] table removed to simplify code, less MEM, little more CPU
 */
void ini_norm2(void){
  ULong i1;
  Tnorm2 sum2=0l,n2max=0; int i2, n2hash, n2num=0;
  tbase ib1;
  static struct sbase b1 /* b2=tmp */;
  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  if (vvv && !mpi_i) printf("\n ini_norm2");
  b1.o=0;

  ERR(-1); /* reset errno for ERR(600) */
  /* initialize hash table (for norm2 distribution) */
  for (i1=0l; i1<N2Blen; i1++) n2h[i1]=n2b[i1]=0;

  /* create scfg-range per block table */
  if (vvv&2 && !mpi_i) printf(" ini_b_l1[]");
  if (!b_l1) b_l1=(tbase *)malloc(sizeof(tbase)*mpi_n*pt_n);
  if (!b_l1) ERR(9011);
  for (i2=0;i2<mpi_n*pt_n;i2++) b_l1[i2] = 0; /* clear for debugging */
  for (i2=0;i2<pt_n;i2++) /* b_ofs must be set also for b_len=0 */
     b_l1[mpi_i*pt_n+i2] =
       ((b_len[mpi_i*pt_n+i2])? getl1(b_ofs[mpi_i*pt_n+i2]
                                     -b_ofs[mpi_i*pt_n])
                             : ~(tbase)0 ); /* dummy cfg > max_cfg */
  // b_l1[mpi_i*pt_n+pt_n]= getl1(node_n1-1); /* ??? */
  /* ToDo: MPI_Allgather(sbuf, 100, MPI_INT, rbuf, 100, MPI_INT, comm);
   *   or get from file
   *  AllGather: rbuf[  0.. 99]=proc0.sbuf[0..99]
   *             rbuf[100..199]=proc1.sbuf[0..99]
   *             ...           =procn.sbuf[0..99]
   */
  /* can we use a part of the receiverarray as sendarray? yes v2.27 */
#ifdef CONFIG_MPI
  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  for(i2=mpi_i*pt_n;i2<(mpi_i+1)*pt_n;i2++)
   if (i2<4 || i2>=mpi_n*pt_n-4)
    printf("\n DBGpreAG  %2d %4d %s",mpi_i,i2,ccfg(b_l1[i2]));
   else if (i2==4) printf("\n ...");
  MPI_Allgather(b_l1+mpi_i*pt_n, pt_n*sizeof(tbase), MPI_CHAR,
                b_l1,            pt_n*sizeof(tbase), MPI_CHAR, MPI_COMM_WORLD);
  if (mpi_i==mpi_n-1)
  for(i2=0;i2<mpi_n*pt_n;i2++) 
    if (i2<4 || i2>=mpi_n*pt_n-4)
     printf("\n DBGpostAG %2d %4d %s",mpi_i,i2,ccfg(b_l1[i2]));
    else if (i2==4) printf("\n ...");
#endif

  if (vvv&2) { /* speed test, see b2i() for results */
    ULong j,loops; long t2=0, t1=itime();
    MPI_Barrier(MPI_COMM_WORLD); /* debug */
    if (!mpi_i) printf(" start_b2i_speedtest");
    for (loops=0;t2<1 && loops<100000;loops++) {
      for(i1=0l; i1<node_n1; i1++){  /* l1[] sort for norm2[]??? */
       ib1=getl1((ULong)i1);       /* l1 is the node part of full l1 */
       j=b2i(ib1); if (j!=i1) break;
      }
      if (i1<node_n1) F2("(j=%lu)!=(i1=%lu)",j,i1);
      t2=itime()-t1;
    }
    if (!mpi_i)
      printf("\n b2i_speedtest.t= %s loops= %lu", sclock(t2), loops);
  }
  
  if (vvv&4) /* not needed at the moment */
  { /* create Isingtable for blocking criteria, Idea from Oct07 */
    /* ToDo: also output a num_states_per_IsingErg I1-I2-matrix */
    int j1,                    /* loop-vars */ 
        min0, max0,             /* min/max numPairs */
        ising,                  /* ising erg, may be exotic hash of all Js */ 
        *buf=NULL;              /* buffer for counted pairs (IsingErg) */
#define MaxIsing (B_NUM)    /* should be prime but 4*(pt_n)..4*NN */
    long numIsing[MaxIsing];   /* IsingErg[j1,j2,..] for load balance */
    long minI,maxI,t1=itime();

    MPI_Barrier(MPI_COMM_WORLD); /* debug */
    printf("\n ini_norm2.IsingBlocks:");
    min0=MaxIsing; max0=0;
    for(j1=0;j1<MaxIsing;j1++) numIsing[j1]=0; /* init hash array */
    /* ToDo: hash should leave near J1-Ising-ergs at near hashs */
    if (sym.num_dcorr*CM(2,4,5)>22)
    for(i1=0l; i1<node_n1; i1++){  /* l1[] sort for norm2[]??? */
      ib1=getl1((ULong)i1);
      buf=get_dcorr(ib1,1,buf); /* get dcorr-table, count ud-pairs */
      /* which bonds we take? worst case is spin glas -> add different bonds */
      /* Idea: make a monte carlo test to find a good function */
      /* IsingErg can be arteficcial build from dcorr, ex: j1+j4/2+... */
      /* ToDo: do hashing within get_dcorr? => hashed j1-jmax-Ising erg */
//    ising=((731*buf[ 5]+23*buf[15]+171*buf[25]+3*buf[35]+17*buf[45])/2)%pt_n;
      ising=((9551*buf[ 2]+23*buf[ 6]
             +1327*buf[10]+ 3*buf[14]
             +  17*buf[18]+ 7*buf[22])/2)%pt_n;
      if (min0>ising) min0=ising;
      if (max0<ising) max0=ising;
      if (ising>=MaxIsing) ising=MaxIsing-1;
      /* hash indicees to avoid overflow */
      numIsing[ising]++;
    }
    printf("\n Ising range for node= %2d: %d-%d", mpi_i, min0, max0);
    /* output Ising distribution, ToDo: only min, max, variance? */
    minI=n1; maxI=0;
    for(j1=min0;j1<MaxIsing && j1<=max0;j1++){
      printf("\n node= %2d Ising hash=%2d: %6ld", mpi_i, j1, numIsing[j1]);
      if (numIsing[j1]<minI) minI= numIsing[j1];
      if (numIsing[j1]>maxI) maxI= numIsing[j1];
    }
    printf("\n node= %2d Ising block size: %lu-%lu-%lu (%.0f%%-%.0f%%)"
           "\n ising.t=%s",
            mpi_i, minI, node_n1/pt_n, maxI,
            minI*100./(node_n1/pt_n),
            maxI*100./(node_n1/pt_n), sclock(itime()-t1));
    if (buf) free(buf); /* malloc by get_dcorr() */
  }
  /* ToDo: sum hash tables? */
  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  
  /* ToDo: remove n2, no significant speedup and MPI-Probs */
  /* if (norm2) ToDo: ini sqrt[n2a][n2b]-table when n2a!=n2b often? */
  for(i1=0l; i1<node_n1; i1++){  /* l1[] sort for norm2[]??? */
   ib1=getl1((ULong)i1);       /* l1 is the node part of full l1 */

   b_ifsmallest3(ib1,&b1); /* called only to get norm2 */
   sum2+=b1.norm2; /* sum2 == numconf ! */
   if(b1.norm2 > n2max) n2max=b1.norm2;	/* store maximum orbit length */

   n2hash=(int)(((ULLong)b1.norm2)%N2Blen);
   if (!n2b[n2hash]) { n2b[n2hash]++; n2h[n2hash]=b1.norm2; n2num++; }
              else if (n2h[n2hash]==b1.norm2) n2b[n2hash]++;
              else { n2b[0]++; /* collision */}
  }
  if (err_fatal==600) ERR(-1); /* reset error 600 */
  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  if (vvv) {
    if (!mpi_i) printf("\n orbit lengths (norm2): ");
    for(i1=1; i1<N2Blen; i1++) if (n2b[i1]) {
      if (mpi_i<4 || mpi_i>=mpi_n-4)
      printf("\n node %2d norm2= %15.10g %8d times",
              mpi_i,(double)n2h[i1],n2b[i1]);
      else if (mpi_i==4) printf("\n ...");
    }
    if (n2b[0]) printf("\n hash_collisions= %8d (should be zero)",n2b[0]);
  }

  MPI_Barrier(MPI_COMM_WORLD); /* debug */
  if (mpi_i<4 || mpi_i>=mpi_n-4)
  printf("\n node %2d n2num= %d  n2max= %.10g sum_norm2= %.15g t+= %s",
          mpi_i, n2num, (double)n2max, (double)sum2, sclock(~0l));
  else if (mpi_i==4) printf("\n ...");

  /* does not work for Tnorm2 == double */
  if ( ((ULLong)n2max) >> (sizeof(ATnorm2)/2*8) == 0 )
    printf("\n# spare memory: Change ATnorm2 to short or char.\n");

  MPI_Barrier(MPI_COMM_WORLD); /* debug */
#ifndef CONFIG_MPI
  /* MaxSBase set to 1 to spare Memory but b_getbase will fail v2.33 */
  if ((vvv&128) && !mpi_i) {
    printf("\nini_norm2 vvv&128 (node=0 only):");
    for(i1=0l;i1<node_n1 && i1<x_max;i1++){
      /* show base for debugging purpose */
      b_ifsmallest3(getl1((ULong)i1),&b1); /* different results for S>=1 ??? */
      printf("\n %6d",(int)b1.norm2);
 /*     b_smallest(getl1((ULong)i1),&b1); *
  *     printf(  " %6d",(int)b1.norm2);   */
      b_getbase(getl1((ULong)i1),&b1);
      printf(  " %6d |%.15g>=",(int)b1.norm2,(double)i1);
      for(i2=0l;i2<b1.num;i2++) 
        printf("\n %32s*%s",outphase(b1.is[i2]),ccfg(b1.ib[i2]));
    }
  }
#endif
  return;
}


/* last modifications: 
   1998-148 type tbase j,m substituted by ulong etc.
   Dez98: general tbase-type independent b2i-fkt. + better errorhandling
 */

/* convert base to index, this is good break point for debugging */
ULong b2i_err(tbase cfg){	/* called if error in b2i() occurs */
  ULong j, m;
  static int err=0;
  static ULong m2=1l;    /* not thread safe */
  static struct sbase b; /* static makes it 100x faster, only used if error */
  int blk, i1, i2, iu, id; 
  err++;
  if (err>128) return ~0; /* giving up to report mass of errors */
  if (err<8)
    printf("\nb2i_err:  %s not found (mpi_i=%d)", ccfg(cfg), mpi_i);
  /* if getl1 was called with index_out_of_range, we have wrong bits */
  for (iu=id=i1=0;i1<nn;i1++) { i2=geti(cfg,i1); if(i2&1)iu++; if(i2&2)id++; }
  if (iu!=nu || id!=nd) {
    printf("\nb2i_err %d: scfg has nu,nd= %d %d which is wrong",mpi_i,iu,id);
    ERR(9999);
  }
  blk=scfg2blk(cfg);
  if (blk/pt_n!=mpi_i) {
    printf("\nb2i_err %d: scfg belongs to wrong block (node=%d,thread=%d)",
              mpi_i, blk/pt_n, blk%pt_n);
    ERR(9999);
  }
  /* recompute the index */
  while ((m2<<1)<node_n1) m2<<=1;	/* Feb00+Nov07 */
  for(j=0l,m=m2;m;m>>=1){if((j|m)<node_n1)if( (getl1(j|m)<=cfg) )j|=m;}
  if (!(getl1(j)==cfg)) /* ok, its still not ok */
  {			/* not found */
   b.o=0;
   b_smallest(cfg,&b);if(err<5)printf("\n cfg_sm = %s phase= %s",ccfg(b.scfg),outphase(b.sign));
   b_getbase( cfg,&b);if(err<5)printf("\n cfg_gb = %s phase= %s",ccfg(b.scfg),outphase(b.sign));
   if (b.sign==0) printf("\n phase==0 ??? b2i() is called illegal ");
   cfg=b.scfg;
   for(j=0l,m=m2;m;m>>=1){ if((j|m)<node_n1)if((getl1(j|m)<=cfg))j|=m; }
   if(!(getl1(j)==cfg)){       /* not found again !!! */
     if(err<3){
       if(j> (ULong)0) printf("\nl1[%5lu]=%s",0l ,ccfg(getl1( 0l)));
       if(j> (ULong)1) printf("\nl1[%5lu]=%s",1l ,ccfg(getl1( 1l)));
       if(j-2<node_n1) printf("\nl1[%5lu]=%s",j-2,ccfg(getl1(j-2)));
       if(j-1<node_n1) printf("\nl1[%5lu]=%s",j-1,ccfg(getl1(j-1)));
       if(j  <node_n1) printf("\nl1[%5lu]=%s",j  ,ccfg(getl1(j  )));
       if(j+1<node_n1) printf("\nl1[%5lu]=%s",j+1,ccfg(getl1(j+1)));
       j=node_n1-1;
       if(node_n1 > 0) printf("\nl1[%5lu]=%s",j  ,ccfg(getl1(j  )));
       ERR(400); /* not found, try again */
     }
     for(j=0;j<node_n1;j++) if(getl1(j)==cfg) break; /* l1 not sorted? */
     if(j==node_n1)if(err<10){ ERR(410); err_fatal++; }
   }
  }
  if (getl1(j)==cfg && err<5)
    printf("\n node %2d: now b2i() is ok and returns %lu", mpi_i, (unsigned long)j);
  return (j);
}

/*
  scfg2idx: sym-config to index within node (0..node_n1-1) (old: b2i)
  the routine is short as possible to fit into the cache
   input is spin-configuration (representator)
   output is index to tables
   ToDo:
    - convert more than one index within one call  for better cache performance
    - take 8 highest 8bit as index best_speedup=lg2(n1)/(lg2(n1)-8)=1.36
      no effect for high Sz, because all configs start with 0-bits?
    - l1_rough_idx[0..1023] Oct07 v2.27
      b2i-speedtest getl1+b2i+if nud=33,7  P4-1.3GHz L2=512kB
       search full table           92loops/9s (bits), 94loops/9s (interval)
       search through rough table 147loops/9s
       no search                 4900loops/9s
       lg2(n1=116769)=17bits lg2(Nrough=1024)=10bits speedup=50%
       no speedup for lg2(n1=1e9)=30bits (speedup=30%)
    Idea: reduce rough table for nodes and threads (no mem collisions!)
          as a compromise
    Idea: sort-as-searchtree: 1..10 to { 5,    2, 7,     1,3, 6,8,  ...}
                                         root  2elements 4elements 8elements
           for better caching? not really -> need hardware-search (or hash?)
    Idea: hash table? or IsingErg
      return -i if cfg outside block? -> no, use extra chck+fkt
    ToDo: search thread_l1_block only?, rename to scfg2i
                                
 */
ULong b2i(tbase cfg){ /* (b=base_config=scfg)-to-index < node_n1, rename to scfg2i! */
  ULong j,ja,jb;
  ja=0; j=jb=node_n1-1;
  /* ToDo: block.h b_idx2blk() looks more nice, without +-1 */
  for (;ja<jb;) { j=(ja+jb+1)/2; if (getl1(j)<=cfg) ja=j; else j=jb=j-1; }
  /* price: log2(node_n1)*(2cmp+2add+1shft+1mem+2jmp) */
  if (getl1(j)==cfg) return j;  /* getl1 returns the node_index=0..node_n1-1 */
  return b2i_err(cfg); /* slow error-routine outside the cpu-i-cache */
}

/* scfg2blk: get global block number from sym-config
 */
int scfg2blk(tbase cfg){ /* get global block number from sym-config */
  int j,ja,jb;
  ja=0; j=jb=pt_n*mpi_n-1;
  for (;ja<jb;) { j=(ja+jb+1)/2; if (b_l1[j]<=cfg) ja=j; else j=jb=j-1; }
  /* price: log2(pt_n*mpi_n)*(2cmp+2add+1shft+1mem+2jmp) */
  return j;  /* getl1 returns the node_index=0..node_n1-1 */
}
