#ifndef hilbert_h 	/* include this only once */
#define hilbert_h 1

/* definition of the base and the operators of the Hilbert space */

#include "modeldef.h"
#include "defs.h"
#include "spins.h"

/* definition of the vakuum state */
#define TBASE0 ((tbase)0)

/* ================= inline - functions ======================= */

/* this function is only for better readability */
/* function returns up and down spin of site i, 0=empty, 1=up, ... */
static inline int geti(tbase cfg,int i){   /* -> 0,1=u,2=d,3=du */
#if CM(1,0,0)
  return 1+(int)(1&(cfg>>i));}
#else
  return( (int)3&(cfg>>i>>i) );}
#endif

/* function set up and/or down spin on site i, 0=empty, 1=up, ... */
static inline tbase puti(tbase cfg,int i,int s){   /* -> 0,1=u,2=d,3=du */
#if CM(1,0,0)
  return (cfg & ~((tbase)1<<i)) | ((tbase)(s-1))<<i; }
#else
  return (cfg & ~(((tbase)3)<<(2*i))) | ((tbase)s)<<(2*i);}
#endif
        

/* fast encode ud[] to cfg, 0 1=u 2=d 3=du ud[0]=lowest Dez01 */
static inline void encode_ud(tbase *cfg, t_sites ud[NN], int nn){
  int i; *cfg=TBASE0;
#if CM(1,0,0)
  for(i=nn-1;i>=0;i--) (*cfg) = ((*cfg)<<1) | ((tbase)(ud[i]-1));
#else
  for(i=0;i<nn;i++)   (*cfg) |= ((tbase)ud[i])<<(2*i);
#endif
}

#if CM(0,1,1)
/* fast xor single spin i ud=0: u, ud=1: d, Apr02 */
static inline tbase tbase_xor(tbase cfg, int i, int ud){
#if CM(1,0,0)
  cfg ^= (((tbase)1)<<i);    /* ud not used!? */
#else
  cfg ^= (((tbase)1)<<(2*i+ud));
#endif
  return cfg;
}
#endif

int b_out(struct sbase *b,int len);
void   paini();
ULLong numsymconf();
ULLong numconf();
void ini_norm2(void);
tbase  ll1(ULLong ii); /* only for ll1(0) ??? */
int initmasks(int nn);

ULong b2i(tbase i);  /* (b=base_config=scfg)-to-index rename to scfg! */
int scfg2blk(tbase i);/* get global block number from sym-config */

int zi(tbase cfg,int i);   /* Siz -> -1,0,1 */
int zizj(tbase cfg,int i,int j);   /* Siz*Sjz -> -1,0,1 */
tbase b_getbase(tbase ib1,struct sbase *b);
#if 0  /* static inline */
void get_symconfig(tbase ib1,int ip,tbase *ib2,tphase *ss,int ms);
void get_sudconfig(tbase ib1,tphase s1,tbase *ib2,tphase *ss);
#endif
    
tbase b_smallest(tbase ib1, struct sbase *b);

#if CM(0,1,1)
/*
 * this comes from fermionic sign (-1)^Sum(n(i<j)) for site j
 *   we calculate a product for 2 sites here
 * Is parity base coding dependend? => only mask depends from base code!
 * +1 if even number of 1-bits,  -1 else
 * inline -> code inserted (faster,only -O and C++)
 * static -> function not externaly available    
 */
static inline int parity(tbase1 ib1){
#if NN>CM(64,32,32)
    int i;              /* this works only with class vlint */
    for(i=64;i<CM(nn,2*nn,2*nn);i<<=1) ib1^=ib1>>i; /* not testet ??? */
#endif
#if NN>CM(32,16,16)
    ib1^=ib1>>32;
#endif
    ib1^=ib1>>16;   ib1^=ib1>>8;   ib1^=ib1>>4;    ib1^=ib1>>2;
    ib1^=ib1<<1;    return -((int)(2&ib1)-1);}
#endif /* CM */

/* see spins.c, depends from base coding (code2)
 *   signum between two sites, for one site use p_mask[j][0] ???
 */
extern tbase1 p_mask[NN][NN];

/* return a unique(?) list of counted different pairs = ising energy index
   uu+dd ud+du ..+33 .3+3. .u+.d+u.+d.+3u+3d+u3+d3  (16pairs -> 5sympairs)
   the list is allocated and must be freed after use
   mode==0: output list, free list, return 0
   mode==1: return list
   buf = return space OR NULL for malloc
 */
int *get_dcorr(tbase b1, int mode, int *buf); /* Apr06, v2.25 */
    

#endif
