/*
  linux:
    mpicc -O2 -g -Wall ../utils/mpi_speed.c
    mpirun -np 8 ./mpi_speed 18 30 >mpi_speed.log &
    sleep 12
    ps x -o pid,time,comm | grep mpi_speed | grep -v 00:00:0\
     | awk '{print "taskset -c -p", NR-1, $1,"; renice 19 ",$1}' | sh
    ps x -o pid,psr,nice,pcpu,time,comm | grep mpi_speed | grep -v 00:00:0
   
  see mpi_speed.gpl and at bottom of this file for data!

*/
#include <stdlib.h>
#include <stdio.h>
#include <mpi.h> 
#include <time.h>

int mpi_n, mpi_i,
    mintime=20; /* minimum seconds run time (max. is 2*mintime) */
char *buf1, *buf2;
static int header=0;

int test_sendrecv(unsigned long len){
  unsigned long i1, i2, i3;
  int node1, node2; time_t t1, t2;
  MPI_Status status;
  if (!header) { header=1;
    if (!mpi_i)
    printf("# threads msgsize[B] loops[2^x] time[s]"
           " latency[us] SumBW[MB/s] error\n"); /* BW=bandwidth*/
  }
  i3=8;
  time(&t1);
  for (i2=0,i1=1l<<i3;i1;i1<<=1,i3++){
    for (;i2<i1;i2++) {  /* include shorter runs to spare usertime */
      node1=(mpi_i-(i2%(mpi_n-1)+1)+mpi_n)%mpi_n;
      node2=(mpi_i+(i2%(mpi_n-1)+1)+mpi_n)%mpi_n;
      MPI_Sendrecv(buf1,len,MPI_BYTE,node1,0,
                   buf2,len,MPI_BYTE,node2,0,MPI_COMM_WORLD,&status);
    }
    time(&t2); t2-=t1;
    if (t2>10 && t2<=mintime && !mpi_i) // some alive messages
    printf("# %8d %8lu %6lu %8d %9.2f  %8.2f  %5.1f%%\n",
           mpi_n, len, i3, (int)t2, t2*1.0e6/i1, i1*1.0e-6*len*mpi_n/t2, 100./t2);
    if (t2>mintime) break;
  }
  if (!mpi_i)
  printf("  %8d %8lu %6lu %8d %9.2f  %8.2f  %5.1f%%\n",
           mpi_n, len, i3, (int)t2, t2*1.0e6/i1, i1*1.0e-6*len*mpi_n/t2, 100./t2);
  return 0;
}

int test_alltoall(unsigned long len){
  unsigned long i1, i2, i3;
  int node1, node2; time_t t1, t2;
  MPI_Status status;
  if (!header) { header=1;
    if (!mpi_i)
    printf("# threads msgsize[B] loops[2^x] time[s]"
           " latency[us] SumBW[MB/s] error\n"); /* BW=bandwidth*/
  }
  i3=8;
  time(&t1);
  for (i2=0,i1=1l<<i3;i1;i1<<=1,i3++){
    for (;i2<i1;i2++) {  /* include shorter runs to spare usertime */
      MPI_Alltoall(buf1,len,MPI_BYTE,
                   buf2,len,MPI_BYTE,MPI_COMM_WORLD);
      // i=1..n: MPI_Send(buff+i*len*sizeof(MPI_BYTE),len,MPI_BYTE,i,...      
      // i=1..n: MPI_Recv(buff+i*len*sizeof(MPI_BYTE),len,MPI_BYTE,i,...      
    }
    time(&t2); t2-=t1;
    if (t2>10 && t2<=mintime && !mpi_i)
    printf("# %8d %8lu %6lu %8d %9.2f  %8.2f  %5.1f%%\n",
           mpi_n, len*mpi_n, i3, (int)t2, t2*1.0e6/i1, i1*1.0e-6*len*mpi_n*mpi_n/t2, 100./t2);
    if (t2>mintime) break;
  }
  if (!mpi_i)
  printf("  %8d %8lu %6lu %8d %9.2f  %8.2f  %5.1f%%\n",
           mpi_n, len*mpi_n, i3, (int)t2, t2*1.0e6/i1, i1*1.0e-6*len*mpi_n*mpi_n/t2, 100./t2);
  return 0;
}


int main(int argc, char *argv[]) {
  unsigned long i1, bufsize, bufsize2=20;
  int version, subversion;
  setvbuf(stdout,NULL,_IONBF,0);
  if (MPI_Init(&argc, &argv) != MPI_SUCCESS) exit(1);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_n);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_i);
  MPI_Get_version(&version,&subversion);
  if (argc>=2) bufsize2=atol(argv[1]);
  if (argc>=3) mintime =atoi(argv[2]);
  bufsize = 1l << bufsize2;
  // -- add a dummy loop here for 12s for a taskset call
  if (!mpi_i) {
    printf("# using MPI %d.%d with %d nodes\n", version, subversion, mpi_n);
    printf("# blocksize: 2^%2ld = %12ld (set by 2^arg1)\n", bufsize2, bufsize);
    printf("# mintime[s]:  %2d                (set by arg2)\n", mintime);
  }
  buf1 = (char *)malloc(bufsize); /* replace by numa_alloc_local */
  buf2 = (char *)malloc(bufsize);
  if ((!buf1) || (!buf2)) MPI_Abort(MPI_COMM_WORLD, 1);

  for (i1=0;i1<bufsize;i1++) buf1[i1]=buf2[i1]=(char)i1;

  MPI_Barrier(MPI_COMM_WORLD);
 
#if 1
  if (!mpi_i) printf("# Measure speed of MPI_sendrecv transactions:\n");
  test_sendrecv(bufsize);
  test_sendrecv(bufsize/2);
  test_sendrecv(bufsize/4);
  test_sendrecv(bufsize/8);
  test_sendrecv(bufsize/16);
  test_sendrecv(2);
  test_sendrecv(1);
#endif
#if 1
  if (!mpi_i) printf("# Measure speed of MPI_alltoall transactions:\n");
  test_alltoall(bufsize/mpi_n);
  test_alltoall(bufsize/mpi_n/2);
  test_alltoall(bufsize/mpi_n/4);
  test_alltoall(bufsize/mpi_n/8);
  test_alltoall(bufsize/mpi_n/16);
  test_alltoall(2);
  test_alltoall(1);
#endif

  free(buf2);
  free(buf1);  
  MPI_Finalize();
  return 0;
}

/*

  P4-1.3GHz-linux 15.3us      476MB/s    (2)
  ES45-Tru64-5.1B  2.6us 5%   820MB/s    (2)
                   2.79us    1130MB/s 2% (4)
      overload-2x 26.0us 4%   7.2MB/s 0.2%(8) SR=Sendrecv AA=Alltoall
  GS320-16:    16 12us 860MB/s(128k) (SR) 156us 880MB/s(128k) (AA)  
     runon 0,1  8 12us 409MB/s(128K) (SR)  82us 373MB/s(128k) # psr=0,1,2,3..3
                4  6us 373MB/s(128K) (SR)  23us 440MB/s(64K) (AA)
                2  5us 260MB/s(128K) (SR)  11us 409MB/s(256K) (AA)
  GS1280-32:    2  3.1us 4%   715MB/s 4%  (slow 32CPUs/RAD)
                8  3.9us 3%  1960MB/s(512kB)
                             1560MB/s(1MB) 3% 
               32  6.4us 4%  3930MB/s(32kB)
                             4040MB/s(64kB,128kB)
                             4160MB/s(256kB)
                             3500MB/s(512kB)
                             2500MB/s(1MB) 3%  
  Altix330 numa_alloc_local MPI_via_loopback ToDo: interconnect-topology=star
     # monitor interconnect-traffic: linkstat, dplace?
     MPT-MPI-1.2 (insmod xpmem) MPI_USE_XPMEM=1 MPI_XPMEM_VERBOSE=1 ?
      2 2.09us 1.8GB/s(1M) 1.1GB/s(256K) 1node  numa_alloc_local
      2 1.97us 8G(1M) 17G 15G 13G 12G 8G 5G 3G 2G 1G 300M ... no div psr1,11
      4        1.6GB/s(1M) 1.6GB/s(256K) 4nodes
      8 3.58us 2.5GB/s(1M) 2.7GB/s(256K) 4nodes
     12 4.05us 3.8GB/s(1M) 3.4GB/s(256K) 6nodes numa_alloc_local
     12 4.05us 4.1GB/s(1M) 34GB/s(128K) 6nodes malloc lo
     12 4.05us 4.3GB/s(1M) 32GB/s(128K) 6nodes malloc xpmem
      # MPI_Sendrecv                    MPI_Alltoall
      4 3.2us  4.2GB/s(1M) 16GB/s(128K) 12.4us  9GB/s(1M) 20GB/s(256K) # running spinpack8
      4 3.1us  4.6GB/s(1M) 20GB/s(128K) 12.4us 11GB/s(1M) 23GB/s(512K) # stopped spinpack8 (parallel-fluent-speed*3.5 14.7.08)
     12 3.9us  3.4GB/s(1M) 37GB/s(128K) 42.0us  6GB/s(1M) 26GB/s(512K) # stopped spinpack8
  2DualAMD265:    2  23us 4%   660MB/s 4% (last!)   baldur
  4Xeon3GHz_tina  4 3.3us 4%   307MB/s 4%  Dec07 node004
    2*1G via .1G  2 160us 5%    15MB/s 3%  Dec07 node4+node012
                  4 191us 4%    28MB/s 3%  Dec07 node4+node012+...
                  8 206us 4%    63MB/s 3%  Dec07 node4+node012+...

  ToDo: show scaling speed(threads)


*/
